#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for cscripts
#
# Copyright (C) 2016-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import sys
import gammalib
import cscripts
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import test_cscript
import test_cscaldb
import test_csbkgmodel
import test_csebins
import test_csfindobs
import test_csinfo
import test_cslightcrv
import test_csmodelinfo
import test_csmodelmerge
import test_csmodelselect
import test_csmodelsois
import test_csobs2caldb
import test_csobsdef
import test_csobsinfo
import test_csobsselect
import test_csphagen
import test_csphasecrv
import test_cspull
import test_csresmap
import test_csresspec
import test_csroot2caldb
import test_cssens
import test_csspec
import test_cssrcdetect
import test_cstsdist
import test_cstsmapmerge
import test_cstsmapsplit
import test_csviscube
import test_csworkflow
import test_obsutils
import test_mputils
import test_csiactcopy
import test_csiactdata
import test_csiactobs


# ================== #
# Perform unit tests #
# ================== #
def test(installed=False, debug=False):
    """
    Perform cscripts unit testing.

    Parameters
    ----------
    installed : bool, optional
        Flag indicating whether the script has been installed or not
    debug : bool, optional
        Enables debugging
    """
    # If we have an installed version then create a temporary directory and
    # copy over all information that is needed
    if installed:

        # Create temporary working directory
        import tempfile
        path = tempfile.mkdtemp()
        os.chdir(path)

        # Get test directory
        import inspect
        testdir = inspect.getfile(cscripts.tests)
        dirname = os.path.dirname(testdir)

        # Copy test data in "data" directory
        os.system('cp -r %s %s' % (dirname+'/data', 'data'))

        # Set test data environment variable
        os.environ['TEST_DATA'] = 'data'

    # ... otherwise set the calibration database to the one shipped with the
    # package; we don't need to set the 'TEST_DATA', this is done by the
    # test environment
    else:
        os.environ['CALDB'] = '%s/caldb' % (os.environ['TEST_SRCDIR'])

    # Create a local "pfiles" directory and set PFILES environment variable
    try:
        os.mkdir('pfiles')
    except:
        pass
    os.environ['PFILES'] = 'pfiles'

    # Copy the cscripts parameter files into the "pfiles" directory. For a
    # non-installed test we copy the parameter files from the "cscripts"
    # source directory into the "pfiles" directory, for an installed version
    # we get all parameter files from the "syspfiles" directory. Also make
    # sure that all parameter files are writable.
    if not installed:
        os.system('cp -r %s/cscripts/*.par pfiles/'% (os.environ['TEST_SRCDIR']))
        os.system('chmod u+w pfiles/*')
    else:
        os.system('cp -r %s/syspfiles/*.par pfiles/' % (os.environ['CTOOLS']))
        os.system('chmod u+w pfiles/*')

    # Define list of test suites
    tests = [test_cscript.Test(),
             test_csbkgmodel.Test(),
             test_cscaldb.Test(),
             test_csebins.Test(),
             test_csinfo.Test(),
             test_cslightcrv.Test(),
             test_csmodelinfo.Test(),
             test_csmodelmerge.Test(),
             test_csmodelselect.Test(),
             test_csmodelsois.Test(),
             test_csobs2caldb.Test(),
             test_csobsdef.Test(),
             test_csobsinfo.Test(),
             test_csobsselect.Test(),
             test_csphagen.Test(),
             test_csphasecrv.Test(),
             test_cspull.Test(),
             test_csresmap.Test(),
             test_csresspec.Test(),
             test_csroot2caldb.Test(),
             test_cssens.Test(),
             test_csspec.Test(),
             test_cssrcdetect.Test(),
             test_cstsdist.Test(),
             test_cstsmapmerge.Test(),
             test_cstsmapsplit.Test(),
             test_csviscube.Test(),
             test_csworkflow.Test(),
             test_obsutils.Test(),
             test_mputils.Test()]

    # Append tests for Python 2.6+ (the IACT cscripts depend on the json
    # module which is only available since Python 2.6+)
    ver = sys.version.split()[0]
    if ver >= '2.6.0':

        # Check for VHEFITS environment variable
        if 'VHEFITS' in os.environ:

            # If the environment variable exists then unset it for the test
            # cases. Since Python is executed in a subprocess this will not
            # impact the environment variable in the parent shell.
            del os.environ['VHEFITS']

        # Append IACT script tests to list of test suites
        tests.extend([test_csfindobs.Test(),
                      test_csiactcopy.Test(),
                      test_csiactdata.Test(),
                      test_csiactobs.Test()])

    # Allocate test suite container
    suites = gammalib.GTestSuites('cscripts unit testing')

    # Set test suites and append them to suite container
    for suite in tests:
        suite.set()
        suites.append(suite)

    # Run test suite
    success = suites.run()

    # Save test results
    if not installed:
        suites.save('reports/cscripts.xml')
    else:
        suites.save('cscripts_reports.xml')

    # If debuging is requested then print test suites
    if debug:
        print(suites)

    # Remove temporary direction
    if installed:
        os.system('rm -rf %s' % (path))

    # Raise an exception in case of failure
    if not success:
        raise RuntimeError('At least one error occured during the test.')

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Run tests
    test()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctobssim tool
#
# Copyright (C) 2014-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from cscripts import obsutils
from testing import test


# ============================ #
# Test class for ctobssim tool #
# ============================ #
class Test(test):
    """
    Test class for ctobssim tools
    
    This test class makes unit tests for the ctobssim tool by using it from
    the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctobssim')

        # Append tests
        self.append(self._test_ctobssim_cmd, 'Test ctobssim on command line')
        self.append(self._test_ctobssim_python, 'Test ctobssim from Python')

        # Return
        return

    # Test ctobssim on command line
    def _test_ctobssim_cmd(self):
        """
        Test ctobssim on the command line
        """
        # Set tool name
        ctobssim = self._tool('ctobssim')

        # Setup ctobssim command
        cmd = ctobssim+' inmodel="'+self._model+'" '+ \
                       ' outevents="ctobssim_cmd1.fits"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'" '+ \
                       ' ra=83.63 dec=22.01 rad=2.0'+ \
                       ' tmin="2020-01-01T00:00:00"'+ \
                       ' tmax="2020-01-01T00:05:00"'+ \
                       ' emin=1.0 emax=100.0'+ \
                       ' logfile="ctobssim_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Load counts cube and check content.
        evt = gammalib.GCTAEventList('ctobssim_cmd1.fits')
        self._test_list(evt, 261)

        # Setup ctobssim command
        cmd = ctobssim+' inmodel="model_that_does_not_exist.xml"'+ \
                       ' outevents="ctobssim_cmd2.fits"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'" '+ \
                       ' ra=83.63 dec=22.01 rad=2.0'+ \
                       ' tmin="2020-01-01T00:00:00"'+ \
                       ' tmax="2020-01-01T00:05:00"'+ \
                       ' emin=1.0 emax=100.0'+ \
                       ' logfile="ctobssim_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctobssim --help
        self._check_help(ctobssim)

        # Return
        return

    # Test ctobssim from Python
    def _test_ctobssim_python(self):
        """
        Test ctobssim from Python
        """
        # Allocate ctobssim
        sim = ctools.ctobssim()

        # Check that empty ctobssim tool holds an empty observation
        self.test_value(sim.obs().size(), 0, 'Check number of observations')

        # Check that saving saves an empty model definition file
        sim['outevents'] = 'ctobssim_py0.fits'
        sim['logfile']   = 'ctobssim_py0.log'
        sim.logFileOpen()
        sim.save()
        self.test_assert(not os.path.isfile('ctobssim_py0.fits'),
             'Check that no event list has been created')

        # Check that clearing does not lead to an exception or segfault
        sim.clear()

        # Now set ctobssim parameters
        sim = ctools.ctobssim()
        sim['inmodel']   = self._model
        sim['caldb']     = self._caldb
        sim['irf']       = self._irf
        sim['ra']        = 83.63
        sim['dec']       = 22.01
        sim['rad']       = 2.0
        sim['tmin']      = '2020-01-01T00:00:00'
        sim['tmax']      = '2020-01-01T00:05:00'
        sim['emin']      = 1.0
        sim['emax']      = 100.0
        sim['outevents'] = 'ctobssim_py1.fits'
        sim['logfile']   = 'ctobssim_py1.log'
        sim['chatter']   = 2

        # Run tool
        sim.logFileOpen()
        sim.run()

        # Check content of observation
        self._test_observation(sim)
        self._test_list(sim.obs()[0].events(), 261)

        # Save events
        sim.save()

        # Load counts cube and check content.
        evt = gammalib.GCTAEventList('ctobssim_py1.fits')
        self._test_list(evt, 261)

        # Copy ctobssim tool
        cpy_sim = sim.copy()

        # Retrieve observation and check content of copy
        self._test_observation(cpy_sim)
        self._test_list(cpy_sim.obs()[0].events(), 261)

        # Execute copy of ctobssim tool again, now with a higher chatter
        # level than before
        cpy_sim['outevents'] = 'ctobssim_py2.fits'
        cpy_sim['logfile']   = 'ctobssim_py2.log'
        cpy_sim['chatter']   = 3
        cpy_sim['publish']   = True
        cpy_sim.logFileOpen()  # Needed to get a new log file
        cpy_sim.execute()

        # Load counts cube and check content.
        evt = gammalib.GCTAEventList('ctobssim_py2.fits')
        self._test_list(evt, 261)

        # Set-up observation container
        pnts = [{'ra': 83.63, 'dec': 21.01},
                {'ra': 84.63, 'dec': 22.01}]
        obs = obsutils.set_obs_list(pnts, caldb=self._caldb, irf=self._irf,
                                    duration=300.0, rad=3.0)

        # Set-up ctobssim tool from observation container
        sim = ctools.ctobssim(obs)
        sim['inmodel']   = self._model
        sim['outevents'] = 'ctobssim_py3.xml'
        sim['logfile']   = 'ctobssim_py3.log'
        sim['chatter']   = 3

        # Double maximum event range
        max_rate = sim.max_rate()
        sim.max_rate(2.0*max_rate)
        self.test_value(sim.max_rate(), 2.0*max_rate, 1.0e-3,
                        'Check setting of maximum event rate')

        # Run ctobssim tool
        sim.logFileOpen()
        sim.run()

        # Retrieve observation and check content
        self._test_observation(sim, nobs=2, pnts=pnts)
        self._test_list(sim.obs()[0].events(), 3630)
        self._test_list(sim.obs()[1].events(), 3594)

        # Save events
        sim.save()

        # Load events
        obs = gammalib.GObservations('ctobssim_py3.xml')

        # Retrieve observation and check content
        self._test_list(obs[0].events(), 3630)
        self._test_list(obs[1].events(), 3594)

        # Return
        return

    # Check observation
    def _test_observation(self, obssim, nobs=1,
                          pnts=[{'ra': 83.63, 'dec': 22.01}],
                          ontime=300.0, livetime=294.0):
        """
        Test content of an observation
        
        Parameters
        ----------
        obssim : `~ctools.ctobssim`
            ctobssim instance
        nobs : int
            Number of observations
        pnts : list of dict
            List of pointing dictionaries
        """
        # Test observation container
        self.test_value(obssim.obs().size(), nobs,
                        'Check number of observations')
        for i, obs in enumerate(obssim.obs()):
            pnt = obs.pointing().dir()
            ra  = pnts[i]['ra']
            dec = pnts[i]['dec']
            self.test_assert(obs.instrument() == 'CTA',
                 'Check if observation is CTA observation')
            self.test_value(obs.ontime(), ontime, 1.0e-6,
                 'Check if ontime is %f sec' % ontime)
            self.test_value(obs.livetime(), livetime, 1.0e-6,
                 'Check if livetime is %f sec' % livetime)
            self.test_value(pnt.ra_deg(), ra, 1.0e-6,
                 'Check if pointing Right Ascension is %f deg' % ra)
            self.test_value(pnt.dec_deg(), dec, 1.0e-6,
                 'Check if pointing Declination is %f deg' % dec)

        # Return
        return

    # Check event list
    def _test_list(self, eventlist, nevents):
        """
        Test content of event list
        
        Parameters
        ----------
        eventlist : `~gammalib.GCTAEventList`
            Event list
        nevents : int
            Expected number of events
        """
        # Test event list
        self.test_value(eventlist.size(), nevents, str(nevents)+' elements')
        self.test_value(eventlist.number(), nevents, str(nevents)+' events')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Shows one or several CTA response functions.
#
# Copyright (C) 2014-2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import math
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ================================== #
# Compute Eq. (17) of Li & Ma (1983) #
# ================================== #
def sigma_lima(Non, Noff, alpha=0.2):
    """
    Compute Eq. (17) of Li & Ma (1983).

    Parameters
    ----------
    Non : float
        Number of On counts
    Noff : float
        Number of Off counts
    alpha : float, optional
        Ratio of on-to-off exposure

    Returns
    -------
    sigma : float
        Detection significance in Gaussian sigma
    """
    # Compute sensitivity in Gaussian sigma
    alpha1 = alpha + 1.0
    Ntotal = Non + Noff
    arg1   = Non  / Ntotal
    arg2   = Noff / Ntotal
    term1  = Non  * math.log((alpha1/alpha)*arg1)
    term2  = Noff * math.log(alpha1*arg2)
    sigma  = math.sqrt(2.0 * (term1 + term2))

    # Return sensitivity
    return sigma


# ======================================== #
# Solve Eq. (17) of Li & Ma (1983) for Non #
# ======================================== #
def Non_lima(sigma, Noff, alpha=0.2):
    """
    Solve Eq. (17) of Li & Ma (1983) for Non

    Parameters
    ----------
    sigma : float
        Required detection significance in Gaussian sigma
    Noff : float
        Number of Off counts
    alpha : float, optional
        Ratio of on-to-off exposure

    Returns
    -------
    Non : float
        Number of On counts
    """
    # Set initial guess for Non
    S   = 25.0
    Non = S + alpha*Noff

    # Compute initial sigma estimate
    s     = sigma_lima(Non, Noff, alpha=alpha)
    delta = s - sigma

    # Iterate
    while abs(delta) > 1.0e-6:
        S    *= sigma/s
        Non   = S + alpha*Noff
        s     = sigma_lima(Non, Noff, alpha=alpha)
        delta = s - sigma

    # Return
    return Non


# ======================= #
# Show one effective area #
# ======================= #
def show_one_effective_area(rsp, name, color='r'):
    """
    Show one effective area

    Parameters
    ----------
    rsp : `~gammalib.GCTAResponse`
        Response function
    name : str
        Name of the response function
    color : str, optional
        Color for plot
    """
    # Set to figure 1
    plt.figure(1)

    # Generate logE(TeV) vector
    nbins = 20
    logE  = [-1.7+0.2*i for i in range(nbins)]
    E     = [math.pow(10.0, e) for e in logE]

    # Generate effective area vector
    aeff = [0.0 for i in range(nbins)]
    for i in range(nbins):
        aeff[i] = rsp.aeff()(logE[i])

    # Plot data
    plt.loglog(E, aeff, color+'-', label=name)
    plt.loglog(E, aeff, color+'o')

    # Set axes
    plt.xlabel('Energy (TeV)')
    plt.ylabel(r'Effective area (cm$^{2}$)')

    # Set legend
    plt.legend(loc='lower right')

    # Return
    return


# ======================== #
# Show one background rate #
# ======================== #
def show_one_background_rate(rsp, name, color='r'):
    """
    Show one background rate

    Parameters
    ----------
    rsp : `~gammalib.GCTAResponse`
        Response function
    name : str
        Name of the response function
    color : str, optional
        Color for plot
    """
    # Set to figure 2
    plt.figure(2)

    # Generate logE(TeV) vector
    nbins = 20
    logE  = [-1.7+0.2*i for i in range(nbins)]
    E     = [math.pow(10.0, e) for e in logE]

    # Generate on-axis background rate vector
    bgrate = [0.0 for i in range(nbins)]
    for i in range(nbins):
        bgrate[i] = rsp.background()(logE[i], 0.0, 0.0)
        if bgrate[i] < 1.0e-10:
            bgrate[i] = 0.0

    # Plot data
    plt.loglog(E, bgrate, color+'-', label=name)
    plt.loglog(E, bgrate, color+'o')

    # Set axes
    plt.xlabel('Energy (TeV)')
    plt.ylabel(r'Background rate (counts s$^{-1}$ MeV$^{-1}$ sr$^{-1}$)')

    # Set legend
    plt.legend(loc='lower left')

    # Return
    return


# ==================== #
# Show one sensitivity #
# ==================== #
def show_one_sensitivity(rsp, name, color='r', duration=180000.0, alpha=0.2,
                         sigma=5.0):
    """
    Show one sensitivity

    Parameters
    ----------
    rsp : `~gammalib.GCTAResponse`
        Response function
    name : str
        Name of the response function
    color : str, optional
        Color for plot
    duration : float, optional
        Duration of observation (s)
    alpha : float, optional
        Ratio of on-to-off exposure
    sigma : float, optional
        Required detection significance in Gaussian sigma
    """
    # Set constants
    r68_to_sigma = 0.6624305
    TeV2erg      = 1.0e6 * gammalib.MeV2erg

    # Set to figure 3
    plt.figure(3)

    # Generate logE(TeV) vector
    nbins = 20
    logE  = [-1.7+0.2*i for i in range(nbins)]
    E     = [math.pow(10.0, e) for e in logE]

    # Initialise sensitivity vector
    flux = [0.0 for i in range(nbins)]

    # Loop over all energy bins
    for i in range(nbins):

        # Compute the energy width in MeV
        ewidth = (math.pow(10.0, logE[i]+0.1) - math.pow(10.0, logE[i]-0.1)) * \
                 1.0e6

        # Extract effective area in cm2 from Instrument Response Function
        aeff = rsp.aeff()(logE[i])

        # Extract 68% containment radius in radians from Instrument Response
        # Function.
        # A kluge is used since this information is not directly available,
        # but the "delta_max()" method returns 5 times the "sigma" of the
        # Gaussian PSF. Hence dividing by 5 gives the "sigma", and dividing
        # by the 68% containment radius to sigma conversion factor gives
        # r68.
        r68 = rsp.psf().delta_max(logE[i]) / 5.0 / r68_to_sigma

        # Compute the solid angle of the 68% containment radius
        solidangle = gammalib.twopi * (1.0 - math.cos(r68))

        # Compute the total number of background counts by multiplying the
        # background rate by the observation duration, the energy width and
        # the solid angle
        bgd_counts = rsp.background()(logE[i], 0.0, 0.0) * \
                     duration * ewidth * solidangle

        # The number of Off counts is the total number of background counts
        # divided by the "alpha" parameter
        Noff = bgd_counts/alpha

        # If there are Off counts then compute the corresponding On counts
        # for "sigma" detection significance
        if Noff > 0:

            # Compute On counts using Li & Ma
            Non = Non_lima(sigma, Noff)

            # Get the source counts by subtracting the  background counts
            # under the source. Make sure that the source counts are at least
            # 5% of the background counts (low-energy background systematics)
            # and that there are at least 10 source events (high-energy limit)
            src_counts = Non - bgd_counts
            if src_counts < 0.05*bgd_counts:
                src_counts = 0.05*bgd_counts
            if src_counts < 10:
                src_counts = 10.0

            # Compute flux under the assumption that the source spectrum is
            # a power law with index of -2.6. The flux is the source counts
            # divided by the duration and 68% of the effective area times
            # a conversion factor. The conversion factor is given by
            #
            #                 conv = 1.6021765 * E^2 / F(E)
            #
            # where E is the mean energy of the bin and F(E) is the flux in
            # the bin
            emin    = gammalib.GEnergy(math.pow(10.0, logE[i]-0.1), 'TeV')
            emax    = gammalib.GEnergy(math.pow(10.0, logE[i]+0.1), 'TeV')
            epivot  = gammalib.GEnergy(math.pow(10.0, logE[i]), 'TeV')
            plaw    = gammalib.GModelSpectralPlaw(1.0e-6, -2.6, epivot)
            conv    = TeV2erg*E[i]*E[i]/plaw.flux(emin, emax)
            flux[i] = conv * src_counts / (duration * aeff*0.68)

    # Plot data
    plt.loglog(E, flux, color+'-', label=name)
    plt.loglog(E, flux, color+'o')

    # Set axes
    plt.xlabel('Energy (TeV)')
    plt.ylabel(r'Sensitivity (erg cm$^{-2}$ s$^{-1}$)')

    # Set legend
    plt.legend(loc='upper right')

    # Return
    return


# ==================== #
# Show one sensitivity #
# ==================== #
def show_one_response(rspname, dbname, name, rootdir=None, color='r'):
    """
    Show one response.

    Parameters
    ----------
    rspname : str
        Response name
    dbname : str
        Database name
    name : str
        Name of the response function
    rootdir : str, optional
        Response root directory
    color : str, optional
        Color for plot
    """
    # Set-up calibration database
    caldb = gammalib.GCaldb()
    if rootdir != None:
        caldb.rootdir(rootdir)
    if gammalib.dir_exists(dbname):
        caldb.rootdir(dbname)
    else:
        caldb.open('cta', dbname)

    # Load response function
    rsp = gammalib.GCTAResponseIrf(rspname, caldb)

    # Show effective area
    show_one_effective_area(rsp, name, color=color)

    # Show background rate
    show_one_background_rate(rsp, name, color=color)

    # Show sensitivity
    show_one_sensitivity(rsp, name, color=color)

    # Return
    return


# ==================== #
# Show one sensitivity #
# ==================== #
def plot_response(plotfile):
    """
    Show response

    Parameters
    ----------
    plotfile : str
        Plot filename
    """
    # Create figures
    plt.figure(1)
    plt.title('Effective area')
    plt.figure(2)
    plt.title('Background rate')
    plt.figure(3)
    plt.title('Sensitivity')

    # Set response dictionary
    rsps = [{'dbname':  'prod2',
             'rspname': 'South_0.5h',
             'name':    'Prod2 (30 min)',
             'color':   'b'},
            {'dbname':  'prod2',
             'rspname': 'South_50h',
             'name':    'Prod2 (50 h)',
             'color':   'r'}]

    # Loop over all responses
    for rsp in rsps:

        # Get attributes
        dbname  = rsp['dbname']
        rspname = rsp['rspname']
        color   = rsp['color']
        name    = rsp['name']
        rootdir = rsp.setdefault('rootdir', None)

        # Show response
        show_one_response(rspname, dbname, name, rootdir=rootdir, color=color)

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ============= #
# Show response #
# ============= #
def show_response():
    """
    Show response
    """
    # Set usage string
    usage = 'show_response.py [-p plotfile]'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    _, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Show response
    plot_response(plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show response
    show_response()



#!/usr/bin/env python
# ==========================================================================
# Selects model from a model definition XML file
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools


# =================== #
# csmodelselect class #
# =================== #
class csmodelselect(ctools.csobservation):
    """
    Selects model from a model definition XML file
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)
        # Set name

        # Initialise class members
        self._models = gammalib.GModels()

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Query input parameters
        self['inobs'].filename()
        self['inmodel'].filename()

        # Query hidden parameters
        self['roilimit'].real()
        self['roimargin'].real()
        self['ethres'].real()
        self['fluxlimit'].real()
        self['tslimit'].real()
        self['fit_pos'].boolean()
        self['fit_shape'].boolean()

        # Query ahead output model filename
        if self._read_ahead():
            self['outmodel'].filename()

        # If there are no observations in container then get them from the
        # parameter file
        if self.obs().size() == 0:
            self.obs(self._get_observations(False))

        # Get models
        self._models = gammalib.GModels(self['inmodel'].filename())

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _select_model(self, model, obs):
        """
        Select model

        Parameters
        ----------
        model : `~gammalib.GModel`
            Model
        obs : `~gammalib.GObservations`
            Observation container

        Returns
        -------
        select : bool
            Model selection flag
        """
        # Get selection parameters
        roilimit  = self['roilimit'].real()
        roimargin = self['roimargin'].real()
        ethres    = self['ethres'].real()
        fluxlimit = self['fluxlimit'].real()
        tslimit   = self['tslimit'].real()

        # Initialise selection flag to True
        select = True
        msg    = 'Select by default'

        # If the model has a spatial component then check of overlap
        if hasattr(model, 'spatial'):

            # Determine the model bounding box
            model_bounds = model.spatial().region()

            # Initialise with an unselect model
            select = False
            msg    = 'Exclude since outside any RoI'

            # Loop over all observations and check whether the model falls
            # into one of them
            for o in obs:

                # Get RoI centre and radius. The RoI radius is limited by
                # roilimit. A margin given by roimargin is added to the RoI
                # radius.
                #obs_centre = o.roi().centre().dir() # Segmentation fault
                #obs_radius = o.roi().radius()       # Segmentation fault
                obs_roi    = o.roi()
                obs_centre = obs_roi.centre().dir()
                obs_radius = obs_roi.radius()
                if obs_radius > roilimit:
                    obs_radius = roilimit
                obs_radius += roimargin

                # Set circular sky region
                obs_bounds = gammalib.GSkyRegionCircle(obs_centre, obs_radius)

                # If model overlaps with RoI then signal overlap
                if obs_bounds.overlaps(model_bounds):
                    select = True
                    msg    = 'Select due to overlap with at least one RoI'
                    break

        # If model is selected and if model has a TS value then apply TS
        # selection
        if select and model.has_ts():
            if model.ts() < tslimit:
                select = False
                msg    = 'Exclude since below TS limit (TS=%.3f)' % model.ts()

        # If model is selected and if model is a sky model then apply flux
        # limit selection
        if select and model.classname() == 'GModelSky':
            emin = gammalib.GEnergy(ethres,   'TeV')
            emax = gammalib.GEnergy(1000.0, 'TeV')
            flux = model.spectral().flux(emin, emax)
            if flux < fluxlimit:
                select = False
                msg    = 'Exclude since below flux limit (Flux=%.3e ph/cm2/s)' % flux

        # Log model selection
        self._log_value(gammalib.NORMAL, model.name(), msg)

        # Return selection flag
        return select

    def _set_model_parameters(self, model):
        """
        Set model parameters

        Parameters
        ----------
        model : `~gammalib.GModel`
            Model

        Returns
        -------
        model : `~gammalib.GModel`
            Model with parameter set
        """
        # Get selection parameters
        fit_pos   = self['fit_pos'].boolean()
        fit_shape = self['fit_shape'].boolean()

        # If the model has a spatial component then set the spatial parameters
        if hasattr(model, 'spatial'):

            # Loop over all parameters
            for par in model:

                # Handle position parameters
                if par.name() == 'RA' or par.name() == 'DEC':
                    if fit_pos:
                        par.free()
                    else:
                        par.fix()

                # Handle shape parameters
                elif par.name() == 'Radius' or \
                     par.name() == 'Sigma' or \
                     par.name() == 'Width' or \
                     par.name() == 'PA' or \
                     par.name() == 'MajorRadius' or \
                     par.name() == 'MinorRadius':
                    if fit_shape:
                        par.free()
                    else:
                        par.fix()

        # Return model
        return model


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Initialise empty model container for selected models
        models = gammalib.GModels()

        # Write input observation container into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Write input models into logger
        self._log_models(gammalib.NORMAL, self._models, 'Input model')

        # Write header
        self._log_header1(gammalib.NORMAL, 'Model selection')

        # Loop over model components
        for model in self._models:

            # If model should be selected then set model parameters and append
            # model to the output container
            if self._select_model(model, self.obs()):

                # Set model parameters
                model = self._set_model_parameters(model)

                # Append model
                models.append(model)

        # Copy selected model into selected models
        self._models = models

        # Write selected models into logger
        self._log_models(gammalib.NORMAL, self._models, 'Selected model')

        # Return
        return

    def save(self):
        """ 
        Save model definition XML file
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save models')

        # Get output filename in case it was not read ahead
        outmodel = self['outmodel'].filename()

        # If file exists and clobber flag is false then raise an exception
        if outmodel.exists() and not self._clobber:
            msg = ('Cannot save "'+outmodel.url()+'": File already exists. '
                   'Use parameter clobber=yes to allow overwriting of files.')
            raise RuntimeError(msg)

        # Otherwise log filename and save file
        else:
            # Log filename
            self._log_value(gammalib.NORMAL, 'Model definition XML file',
                                             outmodel.url())

            # Save models
            self._models.save(outmodel)

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csmodelselect(sys.argv)

    # Execute application
    app.execute()



#!/usr/bin/env python
# ==========================================================================
# Creates a calibration database entry for an IACT observation
#
# Copyright (C) 2015-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import os
import datetime
import gammalib
import ctools
from cscripts import calutils


# ================= #
# csobs2caldb class #
# ================= #
class csobs2caldb(ctools.csobservation):
    """
    Creates a calibration database entry for an IACT observation.

    The creation of a calibration database entry is useful for performing
    simulations for current Imaging Air Cherenkov Telescopes (IACTs).
    The class takes an observation definition XML file as input and uses
    one observation to create a calibration database entry. By default
    the first observation will be used, but the user can specify the
    index of any observation using the hidden "index" parameter.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise members
        self._observation = gammalib.GCTAObservation()
        self._mission     = 'cta'
        self._caldb       = 'cta'
        self._outfile     = gammalib.GFilename('irf_file.fits')
        self._base_dir    = ''
        self._cal_dir     = ''
        self._rsp_dir     = ''
        self._caldb_inx   = gammalib.GFits()
        self._irf_fits    = gammalib.GFits()

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile

        Raises:
            ValueError, IndexError, & RuntimeError.
        """
        # Load observation container from "inobs" parameter in case that
        # the observation container is still empty.
        if self.obs().is_empty():
            self.obs(gammalib.GObservations(self['inobs'].filename()))

        # Raise an exception if the observation container is empty
        if self.obs().is_empty():
            raise ValueError('No or empty observation provided for '
                             'parameter "inobs".')

        # Get index of observation to be used
        index = self['index'].integer()

        # Raise an exception if the index is not valid
        if index < 0 or index >= self.obs().size():
            raise IndexError('Parameter "index=%d" outside the validity '
                             'range [0,%d].' % (index, self.obs().size()))

        # Get observation
        self._observation = self.obs()[index]

        # Make sure we have a CTA observation
        if not self._observation.classname() == 'GCTAObservation':
            raise RuntimeError('Input observation not of type '
                               '"GCTAObservation".')

        # Make sure we have an IRF response associated with the CTA
        # observation
        if not self._observation.response().classname() == 'GCTAResponseIrf':
            raise RuntimeError('Response of input observation not of '
                               'type "GCTAResponseIrf".')

        # Get calibration database name. If the "caldb" parameter is "NONE"
        # or empty then use the instrument name from the observation as
        # calibration database name.
        self._caldb = self['caldb'].string()
        if (self._caldb == 'NONE' or len(self._caldb) == 0):
            self._caldb = self._observation.instrument().lower()

        # Get instrument response function file name
        self._outfile = self['outfile'].filename()

        # Make sure that remaining user parameters are queried now. We
        # do not store the actual parameter values as we do not want
        # too many instance attributes with enhances the maintenance
        # costs.
        self['irf'].string()
        self['rootdir'].string()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _make_irf_file(self):
        """
        Creates an IRF FITS file
        """
        # Write header into logger
        self._log_header2(gammalib.TERSE, 'Creating IRF file')

        # Get response for the observation
        rsp = self._observation.response()

        # Extract response file names
        fname_aeff  = rsp.aeff().filename()
        fname_psf   = rsp.psf().filename()
        fname_edisp = rsp.edisp().filename()
        fname_bkg   = rsp.background().filename()

        # Log filenames
        self._log_header3(gammalib.NORMAL, 'IRF input files')
        self._log_value(gammalib.NORMAL, 'Effective area', fname_aeff.url())
        self._log_value(gammalib.NORMAL, 'Point spread function', fname_psf.url())
        self._log_value(gammalib.NORMAL, 'Energy dispersion', fname_edisp.url())
        self._log_value(gammalib.NORMAL, 'Background rate', fname_bkg.url())

        # Open FITS files of response components
        fits_aeff  = gammalib.GFits(fname_aeff)
        fits_psf   = gammalib.GFits(fname_psf)
        fits_edisp = gammalib.GFits(fname_edisp)
        fits_bkg   = gammalib.GFits(fname_bkg)

        # Get extension names
        ext_aeff  = fname_aeff.extname("EFFECTIVE AREA")
        ext_psf   = fname_psf.extname("POINT SPREAD FUNCTION")
        ext_edisp = fname_edisp.extname("ENERGY DISPERSION")
        ext_bkg   = fname_bkg.extname("BACKGROUND")

        # Create empty FITS file
        fits = gammalib.GFits()

        # Append IRF component to FITS file
        fits.append(fits_aeff[ext_aeff])
        fits.append(fits_psf[ext_psf])
        fits.append(fits_edisp[ext_edisp])
        fits.append(fits_bkg[ext_bkg])

        # Log resulting FITS file
        if self._logNormal():
            self._log(str(fits))
            self._log("\n")
        if self._logExplicit():
            self._log(str(fits[ext_aeff].header()))
            self._log("\n")
            self._log(str(fits[ext_psf].header()))
            self._log("\n")
            self._log(str(fits[ext_edisp].header()))
            self._log("\n")
            self._log(str(fits[ext_bkg].header()))
            self._log("\n")

        # Return fits file
        return fits

    def _make_caldb_index(self):
        """
        Creates an IRF index FITS file
        """
        # Write header into logger
        self._log_header2(gammalib.TERSE, 'Creating database index file')

        # Open calibration database index
        indx_file = self._base_dir + '/caldb.indx'

        # Open index file (or create one if it does not exist)
        cif = gammalib.GFits(indx_file, True)

        # Retrieve "CIF" table
        if cif.contains('CIF'):
            table = cif.table('CIF')

        # ... or create binary table if no "CIF" table exists yet, append
        # binary table to CIF file and return a reference to the appended
        # table to work with
        else:
            table = cif.append(calutils.create_cif_table())

        # Check if output config already exist
        has_config = False
        row_index = -1
        for caldir in table['CAL_DIR']:
            row_index += 1
            if caldir == self._cal_dir:
                has_config = True
                break

        # Create columns if not available
        if not has_config:
            # Append 4 rows to CIF extension
            table.append_rows(4)
            row_index = table.nrows()
        else:
            row_index += 4

        # Add generic information for these 4 rows
        for i in range(4):

            # Set row number
            row = i + row_index-4

            # Set date
            now = str(datetime.datetime.now())
            today, time = now.split()

            # Set element
            table['TELESCOP'][row] = self._mission
            table['INSTRUME'][row] = self._caldb
            table['DETNAM'][row]   = 'NONE'
            table['FILTER'][row]   = 'NONE'
            table['CAL_DEV'][row]  = 'ONLINE'
            table['CAL_CLAS'][row] = 'BCF'
            table['CAL_DTYP'][row] = 'DATA'
            table['CAL_VSD'][row]  = today
            table['CAL_VST'][row]  = time.split('.')[0]
            table['REF_TIME'][row] = 51544.0
            table['CAL_QUAL'][row] = 0
            table['CAL_CBD'][row]  = 'NAME('+self['irf'].string()+')'
            table['CAL_DATE'][row] = today.replace('-','/')[2:]
            table['CAL_DIR'][row]  = self._cal_dir
            table['CAL_FILE'][row] = self._outfile

        # Add effective area information
        row = row_index-4
        table['CAL_CNAM'][row] = 'EFF_AREA'
        table['CAL_DESC'][row] = self._caldb+' effective area'

        # Add point spread function information
        row = row_index-3
        table['CAL_CNAM'][row] = 'RPSF'
        table['CAL_DESC'][row] = self._caldb+' point spread function'

        # Add energy dispersion information
        row = row_index-2
        table['CAL_CNAM'][row] = 'EDISP'
        table['CAL_DESC'][row] = self._caldb+' energy dispersion'

        # Add background information
        row = row_index-1
        table['CAL_CNAM'][row] = 'BKG'
        table['CAL_DESC'][row] = self._caldb+' background'

        # Log resulting FITS table and header
        if self._logNormal():
            self._log(str(table))
            self._log('\n')
        if self._logExplicit():
            self._log(str(table.header()))
            self._log('\n')

        # Return CIF FITS file
        return cif

    def _make_dirs(self):
        """
        Make CALDB directories
        """
        # Write header into logger
        self._log_header2(gammalib.TERSE, 'Creating directory structure')

        # Create calibration database        
        caldb = gammalib.GCaldb(self['rootdir'].string())

        # Set calibration directory 
        self._cal_dir  = 'data'
        self._cal_dir += '/'+self._mission.lower()
        self._cal_dir += '/'+self._caldb.lower()
        self._cal_dir += '/bcf/'+self['irf'].string()

        # Set absolute path
        self._base_dir = caldb.rootdir() +'/data'
        self._base_dir += '/'+self._mission.lower()
        self._base_dir += '/'+self._caldb.lower()

        # Set directory for irf file
        self._rsp_dir = caldb.rootdir() + '/' + self._cal_dir

        # Log resulting FITS table
        self._log_value(gammalib.NORMAL, 'Calibration directory', self._cal_dir)
        self._log_value(gammalib.NORMAL, 'Base directory', self._base_dir)
        if not os.path.isdir(self._rsp_dir):
            name = 'IRF directory'
        else:
            name = 'IRF directory (existing)'
        self._log_value(gammalib.NORMAL, name, self._rsp_dir)

        # Create IRF directory is it does not yet exist
        if not os.path.isdir(self._rsp_dir):
            os.makedirs(self._rsp_dir)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write input parameters and header into logger
        self._log_header1(gammalib.TERSE, 'Creating CALDB entry')

        # Create directory structure
        self._make_dirs()

        # Create/update calibration database   
        self._caldb_inx = self._make_caldb_index()

        # Create response file
        self._irf_fits = self._make_irf_file()

        # Return
        return

    def save(self):
        """
        Save calibration database FITS file
        """
        # Write header into logger
        self._log_header1(gammalib.TERSE, 'Save calibration database')

        # Set response filename
        filename = self._rsp_dir + '/' + self._outfile

        # Write filenames into logger
        self._log_value(gammalib.NORMAL, 'CALDB index file',
                        self._caldb_inx.filename().url())
        self._log_value(gammalib.NORMAL, 'Response file', filename)

        # Save caldb index file
        self._caldb_inx.save(self._clobber())

        # Save response file
        self._irf_fits.saveto(filename, self._clobber())

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csobs2caldb(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cssrcdetect script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================= #
# Test class for cssrcdetect script #
# ================================= #
class Test(test):
    """
    Test class for cssrcdetect script

    This test class makes unit tests for the cssrcdetect script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._map = self._datadir + '/skymap.fits'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('cssrcdetect')

        # Append tests
        self.append(self._test_cmd, 'Test cssrcdetect on command line')
        self.append(self._test_python, 'Test cssrcdetect from Python')

        # Return
        return

    # Test cssrcdetect on command line
    def _test_cmd(self):
        """
        Test cssrcdetect on the command line
        """
        # Set script name
        cssrcdetect = self._script('cssrcdetect')

        # Setup cssrcdetect command
        cmd = cssrcdetect+' inmap="'+self._map+'"'+ \
                          ' outmodel="cssrcdetect_cmd1.xml"'+ \
                          ' outds9file="cssrcdetect_cmd1.reg"'+ \
                          ' corr_rad=0.1'+ \
                          ' srcmodel="POINT" bkgmodel="NONE" threshold=8.0'+ \
                          ' logfile="cssrcdetect_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check model file
        self._check_model_file('cssrcdetect_cmd1.xml', 2)

        # Setup cssrcdetect command
        cmd = cssrcdetect+' inmap="skymap_that_does_not_exist"'+ \
                          ' outmodel="cssrcdetect_cmd1.xml"'+ \
                          ' outds9file="cssrcdetect_cmd1.reg"'+ \
                          ' corr_rad=0.1'+ \
                          ' srcmodel="POINT" bkgmodel="NONE" threshold=8.0'+ \
                          ' logfile="cssrcdetect_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check cssrcdetect --help
        self._check_help(cssrcdetect)

        # Return
        return

    # Test cssrcdetect from Python
    def _test_python(self):
        """
        Test cssrcdetect from Python
        """
        # Allocate empty cssrcdetect script
        srcdetect = cscripts.cssrcdetect()

        # Check that saving does not nothing
        srcdetect['outmodel']   = 'cssrcdetect_py0.xml'
        srcdetect['outds9file'] = 'cssrcdetect_py0.reg'
        srcdetect['logfile']    = 'cssrcdetect_py0.log'
        srcdetect.logFileOpen()
        srcdetect.save()

        # Check than an empty model definition file has been created
        self._check_model_file('cssrcdetect_py0.xml', 0)

        # Check that clearing does not lead to an exception or segfault
        srcdetect.clear()

        # Now set cssrcdetect parameters
        srcdetect['inmap']      = self._map
        srcdetect['outmodel']   = 'cssrcdetect_py1.xml'
        srcdetect['outds9file'] = 'cssrcdetect_py1.reg'
        srcdetect['srcmodel']   = 'POINT'
        srcdetect['bkgmodel']   = 'NONE'
        srcdetect['threshold']  = 8.0
        srcdetect['corr_rad']   = 0.1
        srcdetect['logfile']    = 'cssrcdetect_py1.log'
        srcdetect['chatter']    = 4

        # Run cssrcdetect script and save models
        srcdetect.logFileOpen()   # Make sure we get a log file
        srcdetect.run()
        srcdetect.save()

        # Check model file
        self._check_model_file('cssrcdetect_py1.xml', 2)

        # Execute cssrcdetect script again, now with a higher chatter level
        # than before
        srcdetect['outmodel']   = 'cssrcdetect_py2.xml'
        srcdetect['outds9file'] = 'cssrcdetect_py2.reg'
        srcdetect['bkgmodel']   = 'IRF'
        srcdetect['threshold']  = 8.0
        srcdetect['logfile']    = 'cssrcdetect_py2.log'
        srcdetect['chatter']    = 3
        srcdetect.logFileOpen()  # Needed to get a new log file
        srcdetect.execute()

        # Check model file
        self._check_model_file('cssrcdetect_py2.xml', 3)

        # Return
        return

    # Check model file
    def _check_model_file(self, filename, number):
        """
        Check model file
        """
        # Open models
        models = gammalib.GModels(filename)

        # Check number of models
        self.test_value(models.size(), number, 'Check for '+str(number)+
                        ' models in XML file')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cstsmapsplit script.
#
# Copyright (C) 2016-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================== #
# Test class for cstsmapsplit script #
# ================================== #
class Test(test):
    """
    Test class for cstsmapsplit script

    This test class makes unit tests for the cstsmapsplit script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('cstsmapsplit')

        # Append tests
        self.append(self._test_cmd, 'Test cstsmapsplit on command line')
        self.append(self._test_python, 'Test cstsmapsplit from Python')
        self.append(self._test_pickeling, 'Test cstsmapsplit pickeling')

        # Return
        return

    # Test cstsmapsplit on command line
    def _test_cmd(self):
        """
        Test cstsmapsplit on the command line
        """
        # Set script name
        cstsmapsplit = self._script('cstsmapsplit')

        # Setup cstsmapsplit command
        cmd = cstsmapsplit+' inobs="'+self._events+'"'+ \
                           ' inmodel="'+self._model+'"'+ \
                           ' srcname="Crab"'+ \
                           ' caldb="'+self._caldb+'"'+ \
                           ' irf="'+self._irf+'"'+ \
                           ' outmap="tsmap.fits"'+ \
                           ' nxpix=5'+ \
                           ' nypix=5'+ \
                           ' binsz=0.05'+ \
                           ' coordsys="CEL"'+ \
                           ' xref=83.63'+ \
                           ' yref=22.01'+ \
                           ' proj="CAR"'+ \
                           ' compute_null=yes'+ \
                           ' bins_per_job=5'+ \
                           ' outfile="cstsmapsplit_cmd1.dat"'+ \
                           ' logfile="cstsmapsplit_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check command file
        self._check_cmdfile('cstsmapsplit_cmd1.dat')

        # Setup cstsmapsplit command
        cmd = cstsmapsplit+' inobs="'+self._events+'"'+ \
                           ' inmodel="'+self._model+'"'+ \
                           ' srcname="Does_not_exist"'+ \
                           ' caldb="'+self._caldb+'"'+ \
                           ' irf="'+self._irf+'"'+ \
                           ' outmap="tsmap.fits"'+ \
                           ' nxpix=5'+ \
                           ' nypix=5'+ \
                           ' binsz=0.05'+ \
                           ' coordsys="CEL"'+ \
                           ' xref=83.63'+ \
                           ' yref=22.01'+ \
                           ' proj="CAR"'+ \
                           ' compute_null=yes'+ \
                           ' bins_per_job=5'+ \
                           ' outfile="cstsmapsplit_cmd2.dat"'+ \
                           ' logfile="cstsmapsplit_cmd2.log" debug=yes chatter=2'
        
        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check cstsmapsplit --help
        self._check_help(cstsmapsplit)

        # Return
        return

    # Test cstsmapsplit from Python
    def _test_python(self):
        """
        Test cstsmapsplit from Python
        """
        # Set-up cstsmapsplit
        tsmapsplit = cscripts.cstsmapsplit()
        tsmapsplit['inobs']        = self._events
        tsmapsplit['inmodel']      = self._model
        tsmapsplit['srcname']      = 'Crab'
        tsmapsplit['caldb']        = self._caldb
        tsmapsplit['irf']          = self._irf
        tsmapsplit['outmap']       = 'tsmap.fits'
        tsmapsplit['bins_per_job'] = 5
        tsmapsplit['compute_null'] = False
        tsmapsplit['outfile']      = 'cstsmapsplit_py1.dat'
        tsmapsplit['nxpix']        = 5
        tsmapsplit['nypix']        = 5
        tsmapsplit['binsz']        = 0.05
        tsmapsplit['coordsys']     = 'CEL'
        tsmapsplit['xref']         = 83.6331
        tsmapsplit['yref']         = 22.01
        tsmapsplit['proj']         = 'CAR'
        tsmapsplit['logfile']      = 'cstsmapsplit_py1.log'
        tsmapsplit['chatter']      = 2

        # Run cstsmapsplit script and save the commands
        tsmapsplit.logFileOpen()   # Make sure we get a log file
        tsmapsplit.run()
        tsmapsplit.save()

        # Check command file
        self._check_cmdfile('cstsmapsplit_py1.dat')

        # Test with higher chatter level 3
        tsmapsplit['outfile'] = 'cstsmapsplit_py2.dat'
        tsmapsplit['logfile'] = 'cstsmapsplit_py2.log'
        tsmapsplit['chatter'] = 3
        tsmapsplit.execute()

        # Check command file
        self._check_cmdfile('cstsmapsplit_py2.dat')

        # Test with higher chatter level 4
        tsmapsplit['outfile'] = 'cstsmapsplit_py3.dat'
        tsmapsplit['logfile'] = 'cstsmapsplit_py3.log'
        tsmapsplit['chatter'] = 4
        tsmapsplit.execute()

        # Check command file
        self._check_cmdfile('cstsmapsplit_py3.dat')

        # Return
        return

    # Test cstsmapsplit pickeling
    def _test_pickeling(self):
        """
        Test cstsmapsplit pickeling
        """
        # Perform pickeling tests of empty class
        self._pickeling(cscripts.cstsmapsplit())

        # Set-up cstsmapsplit
        tsmapsplit = cscripts.cstsmapsplit()
        tsmapsplit['inobs']        = self._events
        tsmapsplit['inmodel']      = self._model
        tsmapsplit['srcname']      = 'Crab'
        tsmapsplit['caldb']        = self._caldb
        tsmapsplit['irf']          = self._irf
        tsmapsplit['outmap']       = 'tsmap.fits'
        tsmapsplit['bins_per_job'] = 5
        tsmapsplit['compute_null'] = False
        tsmapsplit['outfile']      = 'cstsmapsplit_py1_pickle.dat'
        tsmapsplit['nxpix']        = 5
        tsmapsplit['nypix']        = 5
        tsmapsplit['binsz']        = 0.05
        tsmapsplit['coordsys']     = 'CEL'
        tsmapsplit['xref']         = 83.6331
        tsmapsplit['yref']         = 22.01
        tsmapsplit['proj']         = 'CAR'
        tsmapsplit['logfile']      = 'cstsmapsplit_py1_pickle.log'
        tsmapsplit['chatter']      = 2

        # Perform pickeling tests of filled class
        obj = self._pickeling(tsmapsplit)

        # Run cstsmapsplit script and save the commands
        obj.logFileOpen()   # Make sure we get a log file
        obj.run()
        obj.save()

        # Return
        return

    # Check observation definition XML file
    def _check_cmdfile(self, filename):
        """
        Check ASCII files containing commands
        """
        # Create file name instance
        fname = gammalib.GFilename(filename)
        
        # Check if execution was successful
        self.test_assert(fname.exists(), 'Check of output filename exists')
        
        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cttsmap tool
#
# Copyright (C) 2014-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# =========================== #
# Test class for cttsmap tool #
# =========================== #
class Test(test):
    """
    Test class for cttsmap tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('cttsmap')

        # Append tests
        self.append(self._test_cmd, 'Test cttsmap on command line')
        self.append(self._test_python, 'Test cttsmap from Python')

        # Return
        return

    # Test cttsmap on command line
    def _test_cmd(self):
        """
        Test cttsmap on the command line
        """
        # Set tool name
        cttsmap = self._tool('cttsmap')

        # Setup cttsmap command
        cmd = cttsmap+' inobs="'+self._events+'"'+ \
                      ' inmodel="'+self._model+'" srcname="Crab"'+ \
                      ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                      ' outmap="cttsmap_cmd1.fits"'+ \
                      ' nxpix=3 nypix=3 binsz=0.02'+ \
                      ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                      ' logfile="cttsmap_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('cttsmap_cmd1.fits')

        # Setup cttsmap command
        cmd = cttsmap+' inobs="event_file_that_does_not_exist.fits"'+ \
                      ' inmodel="'+self._model+'" srcname="Crab"'+ \
                      ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                      ' outmap="cttsmap_cmd2.fits"'+ \
                      ' nxpix=3 nypix=3 binsz=0.02'+ \
                      ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                      ' logfile="cttsmap_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Setup cttsmap command
        cmd = cttsmap+' inobs="'+self._events+'"'+ \
                      ' inmodel="'+self._model+'" srcname="Crab"'+ \
                      ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                      ' errors=yes'+ \
                      ' outmap="cttsmap_cmd3.fits"'+ \
                      ' nxpix=3 nypix=3 binsz=0.02'+ \
                      ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                      ' logfile="cttsmap_cmd3.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('cttsmap_cmd3.fits')

        # Check cttsmap --help
        self._check_help(cttsmap)

        # Return
        return

    # Test cttsmap from Python
    def _test_python(self):
        """
        Test cttsmap from Python
        """
        # Allocate empty cttsmap tool
        tsmap = ctools.cttsmap()

        # Check that empty cttsmap tool holds an empty observation and an
        # empty TS map
        self._check_obs(tsmap.obs(), nobs=0)
        self._check_ts_map(tsmap.tsmap(), nx=0, ny=0)

        # Check that saving does not nothing
        tsmap['outmap']  = 'cttsmap_py0.fits'
        tsmap['logfile'] = 'cttsmap_py0.log'
        tsmap.logFileOpen()
        tsmap.save()
        self.test_assert(not os.path.isfile('cttsmap_py0.fits'),
             'Check that no TS map has been created')

        # Check that clearing does not lead to an exception or segfault
        tsmap.clear()

        # Now set cttsmap parameters
        tsmap['inobs']    = self._events
        tsmap['inmodel']  = self._model
        tsmap['srcname']  = 'Crab'
        tsmap['caldb']    = self._caldb
        tsmap['irf']      = self._irf
        tsmap['nxpix']    = 3
        tsmap['nypix']    = 3
        tsmap['binsz']    = 0.02
        tsmap['coordsys'] = 'CEL'
        tsmap['proj']     = 'CAR'
        tsmap['xref']     = 83.63
        tsmap['yref']     = 22.01
        tsmap['outmap']   = 'cttsmap_py1.fits'
        tsmap['logfile']  = 'cttsmap_py1.log'
        tsmap['chatter']  = 2

        # Run cttsmap tool
        tsmap.logFileOpen()   # Make sure we get a log file
        tsmap.run()
        tsmap.save()

        # Check result file
        self._check_result_file('cttsmap_py1.fits')

        # Copy cttsmap tool
        cpy_tsmap = tsmap.copy()

        # Check observation container and TS map of copy
        self._check_obs(cpy_tsmap.obs())
        self._check_ts_map(cpy_tsmap.tsmap())

        # Execute copy of cttsmap tool again, now with a higher chatter
        # level than before
        cpy_tsmap['binmin']  = 0
        cpy_tsmap['binmax']  = 1
        cpy_tsmap['outmap']  = 'cttsmap_py2.fits'
        cpy_tsmap['logfile'] = 'cttsmap_py2.log'
        cpy_tsmap['chatter'] = 3
        cpy_tsmap['publish'] = True
        cpy_tsmap.logFileOpen()  # Needed to get a new log file
        cpy_tsmap.execute()

        # Check result file
        self._check_result_file('cttsmap_py2.fits')

        # Now clear copy of cttsmap tool
        cpy_tsmap.clear()

        # Check that the cleared copy has also cleared the observation
        # container and TS map
        self._check_obs(cpy_tsmap.obs(), nobs=0)
        self._check_ts_map(cpy_tsmap.tsmap(), nx=0, ny=0)

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nx=3, ny=3):
        """
        Check result file

        Parameters
        ----------
        filename : str
            TS map file name
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        """
        # Open result file
        fits = gammalib.GFits(filename)

        # Get HDUs
        ts    = fits['Primary']
        pre   = fits['Prefactor']
        index = fits['Index']

        # Check dimensions
        self.test_value(ts.naxis(), 2, 'Check for TS map dimensions')
        self.test_value(ts.naxes(0), nx, 'Check for TS map pixels in X')
        self.test_value(ts.naxes(1), ny, 'Check for TS map pixels in Y')
        self.test_value(pre.naxis(), 2, 'Check for Prefactor map dimensions')
        self.test_value(pre.naxes(0), nx, 'Check for Prefactor map pixels in X')
        self.test_value(pre.naxes(1), ny, 'Check for Prefactor map pixels in Y')
        self.test_value(index.naxis(), 2, 'Check for Index map dimensions')
        self.test_value(index.naxes(0), nx, 'Check for Index map pixels in X')
        self.test_value(index.naxes(1), ny, 'Check for Index map pixels in Y')

        # Return
        return

    # Check observation
    def _check_obs(self, obs, nobs=1):
        """
        Check observation

        Parameters
        ----------
        obs : `~gammalib.GObservations`
            Models
        nobs : int, optional
            Expected number of observations
        """
        # Check number of observations
        self.test_value(obs.size(), nobs, 'Check number of observations')

        # Return
        return

    # Check TS map
    def _check_ts_map(self, tsmap, nx=3, ny=3):
        """
        Check TS map

        Parameters
        ----------
        tsmap : `~gammalib.GSkyMap`
            TS map
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        """
        # Determine number of maps
        if nx > 0 and ny > 0:
            nmaps = 1
        else:
            nmaps = 0

        # Check dimensions
        self.test_value(tsmap.nmaps(), nmaps, 'Check number of maps')
        self.test_value(tsmap.nx(), nx, 'Check for number of X pixels')
        self.test_value(tsmap.ny(), ny, 'Check for number of Y pixels')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csmodelselect script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# =================================== #
# Test class for csmodelselect script #
# =================================== #
class Test(test):
    """
    Test class for csmodelselect script

    This test class makes unit tests for the csmodelselect script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csmodelselect')

        # Append tests
        self.append(self._test_cmd, 'Test csmodelselect on command line')
        self.append(self._test_python, 'Test csmodelselect from Python')

        # Return
        return

    # Test csmodelselect on command line
    def _test_cmd(self):
        """
        Test csmodelselect on the command line
        """
        # Set script name
        csmodelselect = self._script('csmodelselect')

        # Setup csmodelselect command
        cmd = csmodelselect+' inobs="'+self._events+'"'+ \
                            ' inmodel="'+self._model+'"'+ \
                            ' outmodel="csmodelselect_cmd1.xml"'+ \
                            ' logfile="csmodelselect_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check model file
        self._check_model_file('csmodelselect_cmd1.xml', 2)

        # Setup csmodelselect command
        cmd = csmodelselect+' inobs="'+self._events+'"'+ \
                            ' inmodel="model_that_does_not_exist.xml"'+ \
                            ' outmodel="csmodelselect_cmd2.xml"'+ \
                            ' logfile="csmodelselect_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csmodelselect --help
        self._check_help(csmodelselect)

        # Return
        return

    # Test csmodelselect from Python
    def _test_python(self):
        """
        Test csmodelselect from Python
        """
        # Allocate empty csmodelselect script
        modelselect = cscripts.csmodelselect()

        # Check that saving does not nothing
        modelselect['outmodel'] = 'csmodelselect_py0.xml'
        modelselect['logfile']  = 'csmodelselect_py0.log'
        modelselect.logFileOpen()
        modelselect.save()

        # Check than an empty model definition file has been created
        self._check_model_file('csmodelselect_py0.xml', 0)

        # Check that clearing does not lead to an exception or segfault
        modelselect.clear()

        # Now set csmodelselect parameters
        modelselect['inobs']    = self._events
        modelselect['inmodel']  = self._model
        modelselect['outmodel'] = 'csmodelselect_py1.xml'
        modelselect['logfile']  = 'csmodelselect_py1.log'
        modelselect['chatter']  = 2

        # Run csmodelselect script and save models
        modelselect.logFileOpen()   # Make sure we get a log file
        modelselect.run()
        modelselect.save()

        # Check model file
        self._check_model_file('csmodelselect_py1.xml', 2)

        # Execute csmodelselect script again, now with a higher chatter level
        # than before
        modelselect['outmodel'] = 'csmodelselect_py2.xml'
        modelselect['logfile']  = 'csmodelselect_py2.log'
        modelselect['chatter']  = 3
        modelselect.logFileOpen()  # Needed to get a new log file
        modelselect.execute()

        # Check model file
        self._check_model_file('csmodelselect_py2.xml', 2)

        # Return
        return

    # Check model file
    def _check_model_file(self, filename, number):
        """
        Check model file
        """
        # Open models
        models = gammalib.GModels(filename)

        # Check number of models
        self.test_value(models.size(), number, 'Check for '+str(number)+
                        ' models in XML file')
        
        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Generates spectra as a function of phase
#
# Copyright (C) 2017-2019 Rolf Buehler
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
from cscripts import obsutils
from cscripts import ioutils
from cscripts import mputils


# ================ #
# csphasecrv class #
# ================ #
class csphasecrv(ctools.csobservation):
    """
    Generates spectra in phase bins

    This script computes spectra by performing a maximum likelihood fit
    using :doc:`ctlike` in a series of phase bins for pulsars. The phase bins
    can be either specified in an ASCII file or as an interval divided into
    equally sized phase bins.

    Examples:
            >>> phcrv = csphasecrv()
            >>> phcrv.run()
            >>> ... (querying for parameters) ...
            >>> phcrv = phrv.phasecrv()
                Generates phase fits and retrieves dictionary with best fit models.

            >>> phcrv = csphasecrv(obs)
            >>> phcrv.execute()
            >>> ... (querying for parameters) ...
                Generates phase fits from the observations 
                and saves results in XML files.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise some members. Phases are stored in a nested list
        # [[ph1min,ph1max], [ph2min,ph2max],..]
        self._srcname      = ''
        self._phbins       = [[0.0,1.0]]
        self._onoff        = False
        self._stacked      = False
        self._fits         = gammalib.GFits()
        self._fitmodels    = {}
        self._nthreads     = 0
        self._excl_reg_map = None # Exclusion region map for on/off analysis

        # Return
        return

    # State methods por pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation getstate method to include some members

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base'         : ctools.csobservation.__getstate__(self),
                 'srcname'      : self._srcname,
                 'phbins'       : self._phbins,
                 'stacked'      : self._stacked,
                 'onoff'        : self._onoff,
                 'fits'         : self._fits,
                 'nthreads'     : self._nthreads,
                 'excl_reg_map' : self._excl_reg_map,
                 'fitmodels'    : self._fitmodels}

        # Return pickled dictionary
        return state

    def __setstate__(self, state):
        """
        Extend ctools.csobservation setstate method to include some members

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        ctools.csobservation.__setstate__(self, state['base'])
        self._srcname      = state['srcname']
        self._phbins       = state['phbins']
        self._stacked      = state['stacked']
        self._onoff        = state['onoff']
        self._fits         = state['fits']
        self._nthreads     = state['nthreads']
        self._excl_reg_map = state['excl_reg_map']
        self._fitmodels    = state['fitmodels']

        # Return
        return

    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Setup observations (require response and allow event list, don't
        # allow counts cube)
        self._setup_observations(self.obs(), True, True, False)

        # Set observation statistic
        self._set_obs_statistic(gammalib.toupper(self['statistic'].string()))

        # Set models if there are none in the container
        if self.obs().models().size() == 0:
            self.obs().models(self['inmodel'].filename())

        # Get phase boundaries
        self._phbins = self._create_tbounds()

        # Set On/Off analysis flag and query relevant user parameters
        self._onoff = self._is_onoff()

        # Get source name
        self._srcname = self['srcname'].string()

        # If cube analysis is selected
        # Set stacked analysis flag and query relevant user parameters
        if not self._onoff:
            self._stacked = self._is_stacked()

        # Query the hidden parameters, just in case
        self['edisp'].boolean()

        # Read ahead output parameters
        if self._read_ahead():
            self['outfile'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Set number of processes for multiprocessing
        self._nthreads = mputils.nthreads(self)

        # Return
        return

    def _create_tbounds(self):
        """
        Creates phase bins

        Returns
        -------
        phbins : `[]`
            List of phase bins
        """
        # Initialise Good Time Intervals
        phbins = []

        # Get algorithm to use for defining the time intervals. Possible
        # values are "FILE" or "LIN". This is enforced at
        # parameter file level, hence no checking is needed.
        algorithm = self['phbinalg'].string()

        # If the algorithm is "FILE" then handle a FITS or an ASCII file for
        # the time bin definition
        if algorithm == 'FILE':

            # Get the filename
            filename = self['phbinfile'].filename()

            # Load  ASCII file as CSV file and construct
            # the GTIs from the rows of the CSV file. It is expected that the
            # CSV file has two columns containing the "START" and "STOP"
            # values of the phase bins. No header row is expected.
            csv = gammalib.GCsv(filename)
            for i in range(csv.nrows()):
                phmin = csv.real(i,0)
                phmax = csv.real(i,1)
                phbins.append([phmin,phmax])

        # If the algorithm is "LIN" then use a linear time binning, defined by
        # the "phbins" user parameters
        elif algorithm == 'LIN':

            # Get start and stop time and number of time bins
            nbins    = self['phbins'].integer()

            # Compute time step and setup the GTIs
            ph_step = 1.0 / float(nbins)
            for i in range(nbins):
                phmin = i     * ph_step
                phmax = (i+1) * ph_step
                phbins.append([phmin,phmax])

        # Return Good Time Intervals
        return phbins

    def _get_free_par_names(self):
        """
        Return list of free parameter names

        Returns
        -------
        names : list of str
            List of free parameter names.
        """
        # Initialise empty list of free parameter names
        names = []

        # Collect list of free parameter names
        for par in self.obs().models()[self._srcname]:
            if par.is_free():
                names.append(par.name())

        # Return names
        return names

    def _create_fits_table(self, results):
        """
        Creates FITS binary table containing phase curve results

        Parameters
        ----------
        results : list of dict
            List of result dictionaries

        Returns
        -------
        table : `~gammalib.GFitsBinTable`
            FITS binary table containing phase curve
        """
        # Determine number of rows in FITS table
        nrows = len(results)

        # Create FITS Table with extension "PHASECURVE"
        table = gammalib.GFitsBinTable(nrows)
        table.extname('PHASECURVE')

        # Append phase columns
        ioutils.append_result_column(table, results, 'PHASE_MIN', '', 'phmin')
        ioutils.append_result_column(table, results, 'PHASE_MAX', '', 'phmax')

        # Append parameter columns
        ioutils.append_model_par_column(table, self.obs().models()[self._srcname],
                                        results)

        # Return table
        return table

    def _create_fits(self):
        """
        Create FITS file object from fit results
        """
        # Initialise list of result dictionaries
        results = []

        # Get source parameters
        pars = self._get_free_par_names()

        # Loop over time bins
        for i in range(len(self._phbins)):

            # Get time boundaries
            phmin = self._phbins[i][0]
            phmax = self._phbins[i][1]

            # Initialise result dictionary
            result = {'phmin': phmin,
                      'phmax': phmax,
                      'pars': pars,
                      'values': {}}
            
            # Store fit results
            phname = str(phmin)+'-'+str(phmax)

            # If the model contains the source of interest fill results
            try:
                source = self._fitmodels[phname][self._srcname]
                for par in pars:
                    result['values'][par]      = source[par].value()
                    result['values']['e_'+par] = source[par].error()

            # ... otherwise fills with zeros
            except:
                for par in pars:
                    result['values'][par]      = 0.
                    result['values']['e_'+par] = 0.

            # Append result to list of dictionaries
            results.append(result)
            
        # Create FITS table from results
        table = self._create_fits_table(results)

        # Create FITS file and append FITS table to FITS file
        self._fits = gammalib.GFits()
        self._fits.append(table)
        
        # Return
        return

    def _save_fits(self):
        """
        Saved phase fit results into a fits file
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save phase FITS file')

        # Create FITS file
        self._create_fits()

        # Get file name
        outfile = self['outfile'].filename()

        # Log file name
        self._log_value(gammalib.NORMAL, 'Phase curve file', outfile.url())

        # Save to fits
        self._fits.saveto(outfile, self._clobber())
        
        # Return
        return

    def _save_xml(self):
        """
        Save phase fit results into one XML model each bin
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save phase XML files')

        # Get file name
        outname = self['outfile'].filename()

        # Loop over all phase bins
        for phbin in self._phbins:

            # Set file name for phase bin
            phname        = str(phbin[0])+'-'+str(phbin[1])
            replace_name  = '_phase_'+phname+'.xml'
            outname_phase = (outname.file()).replace('.fits', replace_name)
            outfile       = outname.path() + outname_phase

            # Save XML file
            self._fitmodels[phname].save(outfile)

            # Log file name
            name = 'Phase [%s] file' % phname
            self._log_value(gammalib.NORMAL, name, outfile)
        
        # Return
        return

    def _phase_bin(self,phbin):

        # Write time bin into header
        self._log_header2(gammalib.TERSE, 'PHASE %f - %f' %
                          (phbin[0], phbin[1]))

        # Select events
        select = ctools.ctselect(self.obs().copy())
        select['emin'] = self['emin'].real()
        select['emax'] = self['emax'].real()
        select['tmin'] = 'UNDEFINED'
        select['tmax'] = 'UNDEFINED'
        select['rad'] = 'UNDEFINED'
        select['ra'] = 'UNDEFINED'
        select['dec'] = 'UNDEFINED'
        select['expr'] = 'PHASE>' + str(phbin[0]) + ' && PHASE<' + str(phbin[1])
        select.run()

        # Set phase string
        phstr = str(phbin[0]) + '-' + str(phbin[1])

        # Add phase to observation id
        for i in range(0, select.obs().size()):
            oldid = select.obs()[i].id()
            select.obs()[i].id(oldid + '_' + phstr)
        obs = select.obs()

        # If an On/Off analysis is requested generate the On/Off observations
        if self._onoff:
            obs = obsutils.get_onoff_obs(self, select.obs(), nthreads=1)

        # ... otherwise, if stacked analysis is requested then bin the
        # events and compute the stacked response functions and setup
        # an observation container with a single stacked observation.
        elif self._stacked:
            obs = obsutils.get_stacked_obs(self, select.obs())

        # Header
        self._log_header3(gammalib.EXPLICIT, 'Fitting the data')

        # The On/Off analysis can produce empty observation containers,
        # e.g., when on-axis observations are used. To avoid ctlike asking
        # for a new observation container (or hang, if in interactive mode)
        # we'll run ctlike only if the size is >0
        if obs.size() > 0:

            # Do maximum likelihood model fitting
            like = ctools.ctlike(obs)
            like['edisp']    = self['edisp'].boolean()
            like['nthreads'] = 1  # Avoids OpenMP conflict
            like.run()

            # Renormalize models to phase selection
            # TODO move the scaling from the temporal to the spectral component
            for model in like.obs().models():
                scaled_norm = model['Normalization'].value() / (phbin[1] - phbin[0])
                model['Normalization'].value(scaled_norm)

            # Store fit model
            fitmodels = like.obs().models().copy()

        # ... otherwise we set an empty model container
        else:
            self._log_string(gammalib.TERSE,
                             'PHASE %f - %f: no observations available'
                             ' for fitting' % (phbin[0], phbin[1]))

            # Set empty models container
            fitmodels = gammalib.GModels()

        # Set results
        result = {'phstr': phstr, 'fitmodels': fitmodels}

        # Return results
        return result


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)
        
        # Get parameters
        self._get_parameters()

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Observation')
        
        # Dictionary to save phase fitted models
        self._fitmodels = {}

        # Write header
        self._log_header1(gammalib.TERSE, 'Generate phase curve')

        # If more than a single thread is requested then use multiprocessing
        if self._nthreads > 1:

            # Compute phase bins
            args        = [(self, '_phase_bin', phbin) for phbin in self._phbins]
            poolresults = mputils.process(self._nthreads, mputils.mpfunc, args)

            # Construct results
            for i in range(len(self._phbins)):
                self._fitmodels[poolresults[i][0]['phstr']] = poolresults[i][0]['fitmodels']
                self._log_string(gammalib.TERSE, poolresults[i][1]['log'], False)

        # Otherwise, loop over all phases
        else:
            for phbin in self._phbins:
                result = self._phase_bin(phbin)
                self._fitmodels[result['phstr']] = result['fitmodels']

        # Create FITS file
        self._create_fits()

        # Optionally publish phase curve
        if self['publish'].boolean():
            self.publish()

        # Return
        return

    def save(self):
        """
        Saves results to fits and XML
        """
        # Save results to FITS
        self._save_fits()

        # Save results to XML files (one per phase bin)
        self._save_xml()
        
        # Return
        return

    def publish(self, name=''):
        """
        Publish phase curve

        Parameters
        ----------
        name : str, optional
            Name of phase curve
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Publish phase curve')

        # Continue only if phase curve is valid
        if self._fits.contains('PHASECURVE'):

            # Set default name is user name is empty
            if not name:
                user_name = self._name()
            else:
                user_name = name

            # Log file name
            self._log_value(gammalib.NORMAL, 'Phase curve name', user_name)

            # Publish phase curve
            self._fits.publish('PHASECURVE', user_name)

        # Return
        return

    def phasecurve(self):
        """
        Return dictionary with best fit models
        """
        # Return
        return self._fitmodels

    def exclusion_map(self, object=None):
        """
        Return and optionally set the exclusion regions map

        Parameters
        ----------
        object : `~gammalib.GSkyRegion` or `~gammalib.GSkyMap` or `~gammalib.GFilename`
            Exclusion regions object

        Returns
        -------
        region : `~gammalib.GSkyRegionMap`
            Exclusion regions map
        """
        # If a regions object is provided then set the regions ...
        if object is not None:
            self._excl_reg_map = gammalib.GSkyRegionMap(object)

        # Return
        return self._excl_reg_map

        
# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csphasecrv(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Generates background model
#
# Copyright (C) 2018-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools


# ================ #
# csbkgmodel class #
# ================ #
class csbkgmodel(ctools.csobservation):
    """
    Generates background model
    """
    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Set members
        self._models     = gammalib.GModels()
        self._instrument = ''

        # Return
        return

    # __getstate__ method for pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation getstate method to include some members

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base'       : ctools.csobservation.__getstate__(self),
                 'models'     : self._models,
                 'instrument' : self._instrument}

        # Return pickled dictionary
        return state

    # __setstate__ method for pickling
    def __setstate__(self, state):
        """
        Extend ctools.csobservation setstate method to include some members

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        ctools.csobservation.__setstate__(self, state['base'])
        self._models     = state['models']
        self._instrument = state['instrument']

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Setup observations (require response, allow event list and counts
        # cube)
        self._setup_observations(self.obs(), True, True, True)

        # Set instrument name
        self._instrument = self._get_instrument()

        # Query input parameters
        spatial = self['spatial'].string()
        if spatial == 'GAUSS(E)':
            snumbins = self['snumbins'].integer()
            if snumbins > 1:
                self['smin'].real()
                self['smax'].real()
        if spatial == 'LOOKUP':
            self['slufile'].filename()
        if spatial == 'LOOKUP'   or spatial == 'GAUSS'   or \
           spatial == 'GAUSS(E)' or spatial == 'PROFILE' or \
           spatial == 'POLYNOM':
            self['gradient'].boolean()
        spectral = self['spectral'].string()
        if spectral == 'NODES':
            self._create_energies()
        self['runwise'].boolean()
        self['emin'].real()
        self['emax'].real()
        self['rad'].real()

        # Query ahead output model filename
        if self._read_ahead():
            self['outmodel'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _get_instrument(self):
        """
        Get instrument name

        Extracts the instrument name from the observation container. If there
        are multiple instruments in the container then use the "instrument"
        parameter to set the instrument name.

        Returns
        -------
        instrument : str
            Instrument name
        """
        # Initialise instrument
        instrument = ''

        # Loop over all observations
        for obs in self.obs():
            if instrument == '':
                instrument = obs.instrument()
            elif instrument != obs.instrument():
                instrument = self['instrument'].string()
                break

        # Return instrument
        return instrument

    def _generate_initial_model(self, sigma_min=2.0, sigma_max=1000.0):
        """
        Generate initial background model

        Parameters
        ----------
        sigma_min : float, optional
            Minimum sigma
        sigma_max : float, optional
            Maximum sigma

        Returns
        -------
        model : `~gammalib.GModelData()`
            Background model
        """
        # Handle IRF model
        if self['spatial'].string() == 'IRF':
            epivot   = gammalib.GEnergy(1.0, 'TeV')
            spectral = gammalib.GModelSpectralPlaw(1.0, 0.0, epivot)
            model    = gammalib.GCTAModelIrfBackground(spectral)

        # Handle AEFF model
        elif self['spatial'].string() == 'AEFF':
            epivot   = gammalib.GEnergy(1.0, 'TeV')
            spectral = gammalib.GModelSpectralPlaw(1.0e-13, -2.5, epivot)
            model    = gammalib.GCTAModelAeffBackground(spectral)

        # Handle other models
        else:

            # Handle LOOKUP model
            if self['spatial'].string() == 'LOOKUP':

                # Set spatial model
                factor1 = gammalib.GCTAModelSpatialLookup(self['slufile'].filename())

            # Handle GAUSS model
            elif self['spatial'].string() == 'GAUSS':

                # Set spatial model
                factor1 = gammalib.GCTAModelRadialGauss(3.0)
                factor1['Sigma'].min(sigma_min)
                factor1['Sigma'].max(sigma_max)

            # Handle GAUSS(E) model
            elif self['spatial'].string() == 'GAUSS(E)':

                # Set spatial model. If a single energy bin is specified then
                # use the GAUSS model. Otherwise allocate a node spectrum for
                # the energy nodes of GAUSS(E).
                if self['snumbins'].integer() == 1:
                    factor1 = gammalib.GCTAModelRadialGauss(3.0)
                    factor1['Sigma'].min(sigma_min)
                    factor1['Sigma'].max(sigma_max)
                else:
                    emin     = gammalib.GEnergy(self['smin'].real(), 'TeV')
                    emax     = gammalib.GEnergy(self['smax'].real(), 'TeV')
                    energies = gammalib.GEnergies(self['snumbins'].integer(), emin, emax)
                    spectrum = gammalib.GModelSpectralConst(3.0)
                    nodes    = gammalib.GModelSpectralNodes(spectrum, energies)
                    for i in range(nodes.nodes()):
                        nodes[i*2+1].min(sigma_min)
                        nodes[i*2+1].max(sigma_max)
                    nodes.autoscale()
                    factor1 = gammalib.GCTAModelSpatialGaussSpectrum(nodes)

            # Handle PROFILE model
            elif self['spatial'].string() == 'PROFILE':

                # Set spatial model
                factor1 = gammalib.GCTAModelRadialProfile(2.0, 4.0, 5.0)
                factor1['Width'].min(1.0)
                factor1['Width'].max(10.0)
                factor1['Core'].min(1.0)
                factor1['Core'].max(10.0)
                factor1['Tail'].min(1.0)
                factor1['Tail'].max(10.0)
                factor1['Tail'].free()

            # Handle POLYNOM model
            elif self['spatial'].string() == 'POLYNOM':

                # Set spatial model
                factor1 = gammalib.GCTAModelRadialPolynom([1.0, -0.1, +0.1])
                factor1['Coeff0'].min(0.1)
                factor1['Coeff0'].max(10.0)
                factor1['Coeff0'].fix()
                factor1['Coeff1'].min(-10.0)
                factor1['Coeff1'].max(10.0)
                factor1['Coeff2'].min(-10.0)
                factor1['Coeff2'].max(10.0)

            # Any other strings (should never occur)
            else:
                model = None

            # Optionally add gradient
            if self['gradient'].boolean():
                spatial = gammalib.GCTAModelSpatialMultiplicative()
                factor2 = gammalib.GCTAModelSpatialGradient()
                spatial.append(factor1)
                spatial.append(factor2)
            else:
                spatial = factor1

            # Set spectral model
            epivot   = gammalib.GEnergy(1.0, 'TeV')
            spectral = gammalib.GModelSpectralPlaw(3.0e-4, -1.5, epivot)
            spectral['Prefactor'].min(1.0e-8)

            # Set background model
            model = gammalib.GCTAModelBackground(spatial, spectral)

        # Set background name and instrument
        if model is not None:
            model.name('Background')
            model.instruments(self._instrument)

        # Return model
        return model

    def _select_events(self, obs):
        """
        Select events within a given RoI radius

        Parameters
        ----------
        obs : `~gammalib.GObservations()`
            Observation container

        Returns
        -------
        obs : `~gammalib.GObservations()`
            Observation container
        """
        # Setup task parameters
        select = ctools.ctselect(obs)
        select['ra']       = 'UNDEF'
        select['dec']      = 'UNDEF'
        select['rad']      = self['rad'].real()
        select['tmin']     = 'UNDEF'
        select['tmax']     = 'UNDEF'
        select['emin']     = self['emin'].real()
        select['emax']     = self['emax'].real()
        select['usethres'] = 'DEFAULT'

        # Select events
        select.run()

        # Extract observation
        obs = select.obs().copy()

        # Return observation
        return obs

    def _create_nodes(self, model):
        """
        Replace spectral model by node function

        Parameters
        ----------
        model : `~gammalib.GModelData()`
            Input background model

        Returns
        -------
        model : `~gammalib.GModelData()`
            Background model with spectral node function
        """
        # Extract spectral model of background component
        spectrum = model.spectral()

        # Define node energies
        energies  = self._create_energies()

        # Create node spectrum
        nodes = gammalib.GModelSpectralNodes(spectrum, energies)
        nodes.autoscale()

        # Set minimum node intensities (this avoids NaNs during model fitting)
        for i in range(nodes.size()):
            nodes[i].min(1.0e-10*nodes[i].value())

        # Set node spectrum
        model.spectral(nodes)

        # Return
        return model

    def _generate_bkg(self, obs):
        """
        Generate background models

        Parameters
        ----------
        obs : `~gammalib.GObservations()`
            Observations container

        Returns
        -------
        model : `~gammalib.GModelData()`
            Background model component
        """
        # Write header for event selection
        self._log_header3(gammalib.EXPLICIT, 'Select events from observation')

        # Select events
        obs = self._select_events(obs)

        # Write header for initial background model generation
        self._log_header3(gammalib.EXPLICIT, 'Generate initial background model')

        # Generate initial background model
        model = self._generate_initial_model()

        # Attach initial background model
        models = gammalib.GModels()
        models.append(model)
        obs.models(models)

        # Write header for initial model fitting
        self._log_header3(gammalib.EXPLICIT, 'Fit initial background model')

        # Perform maximum likelihood fitting with initial model
        like = ctools.ctlike(obs)
        like.run()

        # Extract optimiser
        opt = like.opt()

        # Extract fitted model
        model = like.obs().models()[0].copy()

        # If a NODES model is requested then refit a node spectrum
        if self['spectral'].string() == 'NODES':

            # Create nodes spectrum from fitted initial model
            model = self._create_nodes(model)

            # Attach node spectrum
            models = gammalib.GModels()
            models.append(model)
            obs.models(models)

            # Write header for node model fitting
            self._log_header3(gammalib.EXPLICIT, 'Fit nodes background model')

            # Perform maximum likelihood fitting with node model
            like = ctools.ctlike(obs)
            like.run()

            # Extract optimiser
            opt = like.opt()

            # Extract fitted model
            model = like.obs().models()[0].copy()

            # Remove nodes with zero errors as they are not constrained
            # by the data and may lead to fitting problems later
            spectral = model.spectral()
            nodes    = spectral.nodes()
            for i in range(nodes):
                iint = 2*(nodes - i) - 1
                if spectral[iint].error() == 0.0:
                    spectral.remove(nodes - i - 1)

        # Write optimizer
        self._log_string(gammalib.EXPLICIT, str(opt))

        # Return model
        return model

    def _generate_runwise_bkg(self):
        """
        Generate background models
        """
        # Loop over observations
        for run in self.obs():

            # Write header
            self._log_header2(gammalib.TERSE, self._get_obs_header(run))

            # Build observation container with single run
            obs = gammalib.GObservations()
            obs.append(run)

            # Generate background model for
            model = self._generate_bkg(obs)

            # Set model attributes
            model.name(('Background_%s' % run.id()))
            model.ids(run.id())

            # Write model
            self._log_string(gammalib.NORMAL, str(model))

            # Append background model
            self._models.append(model)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Write header
        self._log_header1(gammalib.TERSE, 'Generate background models')

        # Clear models
        self._models.clear()

        # Generate background models
        if self['runwise'].boolean():
            self._generate_runwise_bkg()
        else:
            model = self._generate_bkg(self.obs())
            self._models.append(model)

        # Write models
        self._log_models(gammalib.NORMAL, self._models, 'Output model')

        # Return
        return

    def save(self):
        """
        Save models
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save models')

        # Get models filename
        outmodel = self['outmodel'].filename()

        # Log file name
        self._log_value(gammalib.NORMAL, 'Models file', outmodel.url())

        # Save models
        self._models.save(outmodel)

        # Return
        return

    def models(self):
        """
        Return background models

        Returns
        -------
        ebounds : `~gammalib.GModels()`
            Background models
        """
        # Return
        return self._models


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csbkgmodel(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csviscube script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# =============================== #
# Test class for csviscube script #
# =============================== #
class Test(test):
    """
    Test class for csviscube script
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csviscube')

        # Append tests
        self.append(self._test_cmd, 'Test csviscube on command line')
        self.append(self._test_python, 'Test csviscube from Python')

        # Return
        return

    # Test csviscube on command line
    def _test_cmd(self):
        """
        Test csviscube on the command line
        """
        # Set script name
        csviscube = self._script('csviscube')

        # Setup csviscube command
        cmd = csviscube+' tmin=0.0 tmax=10000.0 geolon=79.4041'+\
                        ' geolat=-24.6272 binsz=10.0 outfile=csviscube_cmd1.fits'+\
                        ' logfile="csviscube_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check visibility cube
        self._check_viscube('csviscube_cmd1.fits')

        # Check csviscube --help
        self._check_help(csviscube)

        # Return
        return

    # Test csviscube from Python
    def _test_python(self):
        """
        Test csviscube from Python
        """
        # Set-up csviscube
        viscube = cscripts.csviscube()
        viscube['tmin']    =  0.0
        viscube['tmax']    =  10000.0
        viscube['geolon']  =  79.4041
        viscube['geolat']  = -24.6272
        viscube['binsz']   = 10.0
        viscube['outfile'] = 'csviscube_py1.fits'
        viscube['logfile'] = 'csviscube_py1.log'
        viscube['chatter'] = 2
        viscube['publish'] = True

        # Execute script
        viscube.logFileOpen()   # Make sure we get a log file
        viscube.execute()

        # Check visibility cube
        self._check_viscube('csviscube_py1.fits')

        # Return
        return

    # Check visibility cube
    def _check_viscube(self, filename, nx=36, ny=18, nz=60):
        """
        Check visibility cube

        Parameters
        ----------
        filename : str
            Visibility cube file name
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        """
        # Open result file
        viscube = gammalib.GSkyMap(filename)

        # Check dimensions
        self.test_value(viscube.nmaps(), nz, 'Check number of maps')
        self.test_value(viscube.nx(), nx, 'Check for number of X pixels')
        self.test_value(viscube.ny(), ny, 'Check for number of Y pixels')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csroot2caldb script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import cscripts
from testing import test


# ================================== #
# Test class for csroot2caldb script #
# ================================== #
class Test(test):
    """
    Test class for csroot2caldb script

    This test class makes unit tests for the cslightcrv script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._infile = self._datadir + '/irf.root'

        # Clean calibration database
        os.system('rm -rf csroot2caldb_* >/dev/null 2>&1')

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csroot2caldb')

        # Append tests
        self.append(self._test_cmd, 'Test csroot2caldb on command line')
        self.append(self._test_python, 'Test csroot2caldb from Python')

        # Return
        return

    # Test csroot2caldb on command line
    def _test_cmd(self):
        """
        Test csroot2caldb on the command line
        """
        # Set script name
        csroot2caldb = self._script('csroot2caldb')

        # Setup csroot2caldb command
        cmd = csroot2caldb+' infile="'+self._infile+'"'+ \
                           ' outdir="csroot2caldb_cmd1"'+ \
                           ' inst=prod3 id=South_50h version=prod3'+ \
                           ' analysis=DESY zenith=20.0 azimuth=0.0'+ \
                           ' logfile="csroot2caldb_cmd1.log" chatter=2'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check csroot2caldb --help
        self._check_help(csroot2caldb)

        # Return
        return

    # Test csroot2caldb from Python
    def _test_python(self):
        """
        Test csroot2caldb from Python
        """
        # Set-up csroot2caldb with a maximum of parameters so that the
        # corresponding code gets tests
        caldb = cscripts.csroot2caldb()
        caldb['infile']        = self._infile
        caldb['outdir']        = 'csroot2caldb_py1'
        caldb['inst']          = 'prod3'
        caldb['id']            = 'South_50h'
        caldb['version']       = 'prod3'
        caldb['analysis']      = 'DESY'
        caldb['zenith']        = 20.0
        caldb['azimuth']       = 0.0
        caldb['psftype']       = 'King'
        caldb['norm1d']        = True
        caldb['rebin']         = True
        caldb['eascale']       = 1.1
        caldb['bgdscale']      = 0.9
        caldb['bgdoversample'] = 3
        caldb['bgdethres']     = 10.0
        caldb['bgdinfill']     = True
        caldb['logfile']       = 'csroot2caldb_py1.log'
        caldb['chatter']       = 4

        # Run script
        caldb.logFileOpen()   # Make sure we get a log file
        caldb.run()

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the obsutils module
#
# Copyright (C) 2017-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from cscripts import obsutils
from testing import test


# ============================== #
# Test class for obsutils module #
# ============================== #
class Test(test):
    """
    Test class for obsutils module
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test data
        self._model = self._datadir + '/model_crab_radialacceptance.xml'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('obsutils')

        # Append tests
        self.append(self._test_sim_unbinned,
                    'Test obsutils.sim() function in unbinned mode')
        self.append(self._test_sim_binned,
                    'Test obsutils.sim() function in binned mode')
        self.append(self._test_sim_stacked,
                    'Test obsutils.sim() function in stacked mode')
        self.append(self._test_sim_stacked_edisp,
                    'Test obsutils.sim() function in stacked mode with energy '
                    'dispersion')
        self.append(self._test_sim_onoff,
                    'Test obsutils.sim() function in On/Off mode')
        self.append(self._test_sim_log,
                    'Test obsutils.sim() function with logging switched on')
        self.append(self._test_set_obs,
                    'Test obsutils.set_obs() function')
        self.append(self._test_set_obs_list,
                    'Test obsutils.set_obs_list() function')
        self.append(self._test_set_obs_patterns,
                    'Test obsutils.set_obs_patterns() function')
        self.append(self._test_set_observations,
                    'Test obsutils.set_observations() function')
        self.append(self._test_get_stacked_response,
                    'Test obsutils.get_stacked_response() function')
        self.append(self._test_get_stacked_obs,
                    'Test obsutils.get_stacked_obs() function')
        self.append(self._test_get_onoff_obs,
                    'Test obsutils.get_onoff_obs() function')

        # Return
        return

    # Setup method for sim() function test
    def _setup_sim(self, two=False):
        """
        Setup method for sim() function test
        """
        # Set-up observation container
        pnt = gammalib.GSkyDir()
        pnt.radec_deg(83.6331, 22.0145)
        obs = gammalib.GObservations()
        run = obsutils.set_obs(pnt, duration=20.0, emin=1.0, emax=10.0)
        run.id('0')
        obs.append(run)
        if two:
            run.id('1')
            obs.append(run)

        # Append model
        obs.models(gammalib.GModels(self._model))

        # Return
        return obs

    # Test sim() function in unbinned mode
    def _test_sim_unbinned(self):
        """
        Test sim() function in unbinned mode
        """
        # Simulate unbinned observations
        res = obsutils.sim(self._setup_sim())

        # Check simulation results
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 4, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'EventList', 'Check event type')
        self.test_value(res[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res[0].events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')
        self.test_value(res[0].events().number(), 4,
                        'Check number of events in list')

        # Check energy dispersion flag
        self.test_assert(not (res[0].response().use_edisp()),
                         'Check energy dispersion usage')
        res[0].response().apply_edisp(True)
        self.test_assert(res[0].response().use_edisp(),
                         'Check energy dispersion usage')

        # Return
        return

    # Test sim() function in binned mode
    def _test_sim_binned(self):
        """
        Test sim() function in binned mode
        """
        # Simulate binned observations
        res = obsutils.sim(self._setup_sim(), nbins=5)

        # Check simulation results
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 4, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'CountsCube', 'Check event type')
        self.test_value(res[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res[0].events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')
        self.test_value(res[0].events().ebounds().size(), 5,
                        'Check number of energy bins')
        self.test_value(res[0].events().number(), 4,
                        'Check number of events in cube')

        # Check energy dispersion flag
        self.test_assert(not (res[0].response().use_edisp()),
                         'Check energy dispersion usage')
        res[0].response().apply_edisp(True)
        self.test_assert(res[0].response().use_edisp(),
                         'Check energy dispersion usage')

        # Return
        return

    # Test sim() function in stacked mode
    def _test_sim_stacked(self):
        """
        Test sim() function in stacked mode
        """
        # Simulate stacked observations
        res = obsutils.sim(self._setup_sim(two=True), nbins=5)

        # Check simulation results
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 15, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'CountsCube', 'Check event type')
        self.test_value(res[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res[0].events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')
        self.test_value(res[0].events().ebounds().size(), 5,
                        'Check number of energy bins')
        self.test_value(res[0].events().number(), 15,
                        'Check number of events in cube')

        # Check energy dispersion flag
        self.test_assert(not (res[0].response().use_edisp()),
                         'Check energy dispersion usage')
        res[0].response().apply_edisp(True)
        self.test_assert(not (res[0].response().use_edisp()),
                         'Check energy dispersion usage')

        # Return
        return

    # Test sim() function in stacked mode with energy dispersion
    def _test_sim_stacked_edisp(self):
        """
        Test sim() function in stacked mode with energy dispersion
        """
        # Simulate binned observations
        res = obsutils.sim(self._setup_sim(two=True), nbins=5, edisp=True,
                           log=True)

        # Check simulation results
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 20, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'CountsCube', 'Check event type')
        self.test_value(res[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res[0].events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')
        self.test_value(res[0].events().ebounds().size(), 5,
                        'Check number of energy bins')
        self.test_value(res[0].events().number(), 20,
                        'Check number of events in cube')

        # Check energy dispersion flag
        self.test_assert(not (res[0].response().use_edisp()),
                         'Check energy dispersion usage')
        res[0].response().apply_edisp(True)
        self.test_assert(res[0].response().use_edisp(),
                         'Check energy dispersion usage')

        # Return
        return

    # Test sim() function in On/Off mode
    def _test_sim_onoff(self):
        """
        Test sim() function in On/Off mode
        """
        # Set-up observation container
        pnt = gammalib.GSkyDir()
        pnt.radec_deg(83.63, 22.51)
        obs = gammalib.GObservations()
        run = obsutils.set_obs(pnt, duration=100.0, emin=1.0, emax=10.0, obsid='0')
        obs.append(run)
        pnt.radec_deg(83.63, 21.51)
        run = obsutils.set_obs(pnt, duration=100.0, emin=1.0, emax=10.0, obsid='1')
        obs.append(run)
        obs.models(gammalib.GModels(self._model))

        # Simulate On/Off observations
        res = obsutils.sim(obs, onsrc='Crab', nbins=5)

        # Check simulation results
        self.test_value(res.size(), 2, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 48, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')

        # Check results of first observation
        self.test_value(res[0].on_spec().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy of On spectrum')
        self.test_value(res[0].on_spec().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy of On spectrum')
        self.test_value(res[0].on_spec().ebounds().size(), 5,
                        'Check number of energy bins of On spectrum')
        self.test_value(res[0].on_spec().counts(), 26,
                        'Check number of events in of On spectrum')
        self.test_value(res[0].off_spec().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy of Off spectrum')
        self.test_value(res[0].off_spec().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy of Off spectrum')
        self.test_value(res[0].off_spec().ebounds().size(), 5,
                        'Check number of energy bins of Off spectrum')
        self.test_value(res[0].off_spec().counts(), 1,
                        'Check number of events in of Off spectrum')
        self.test_value(res[0].arf().ebounds().emin().TeV(), 0.5,
                        'Check minimum energy of ARF')
        self.test_value(res[0].arf().ebounds().emax().TeV(), 12.0,
                        'Check maximum energy of ARF')
        self.test_value(res[0].arf().ebounds().size(), 41,
                        'Check number of energy bins of ARF')
        self.test_value(res[0].rmf().etrue().emin().TeV(), 0.5,
                        'Check minimum true energy of RMF')
        self.test_value(res[0].rmf().etrue().emax().TeV(), 12.0,
                        'Check maximum true energy of RMF')
        self.test_value(res[0].rmf().etrue().size(), 41,
                        'Check number of true energy bins of RMF')
        self.test_value(res[0].rmf().emeasured().emin().TeV(), 1.0,
                        'Check minimum reconstructed energy of RMF')
        self.test_value(res[0].rmf().emeasured().emax().TeV(), 10.0,
                        'Check maximum reconstructed energy of RMF')
        self.test_value(res[0].rmf().emeasured().size(), 5,
                        'Check number of reconstructed energy bins of RMF')

        # Check results of second observation
        self.test_value(res[1].on_spec().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy of On spectrum')
        self.test_value(res[1].on_spec().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy of On spectrum')
        self.test_value(res[1].on_spec().ebounds().size(), 5,
                        'Check number of energy bins of On spectrum')
        self.test_value(res[1].on_spec().counts(), 22,
                        'Check number of events in of On spectrum')
        self.test_value(res[1].off_spec().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy of Off spectrum')
        self.test_value(res[1].off_spec().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy of Off spectrum')
        self.test_value(res[1].off_spec().ebounds().size(), 5,
                        'Check number of energy bins of Off spectrum')
        self.test_value(res[1].off_spec().counts(), 1,
                        'Check number of events in of Off spectrum')
        self.test_value(res[1].arf().ebounds().emin().TeV(), 0.5,
                        'Check minimum energy of ARF')
        self.test_value(res[1].arf().ebounds().emax().TeV(), 12.0,
                        'Check maximum energy of ARF')
        self.test_value(res[1].arf().ebounds().size(), 41,
                        'Check number of energy bins of ARF')
        self.test_value(res[1].rmf().etrue().emin().TeV(), 0.5,
                        'Check minimum true energy of RMF')
        self.test_value(res[1].rmf().etrue().emax().TeV(), 12.0,
                        'Check maximum true energy of RMF')
        self.test_value(res[1].rmf().etrue().size(), 41,
                        'Check number of true energy bins of RMF')
        self.test_value(res[1].rmf().emeasured().emin().TeV(), 1.0,
                        'Check minimum reconstructed energy of RMF')
        self.test_value(res[1].rmf().emeasured().emax().TeV(), 10.0,
                        'Check maximum reconstructed energy of RMF')
        self.test_value(res[1].rmf().emeasured().size(), 5,
                        'Check number of reconstructed energy bins of RMF')

        # Return
        return

    # Test sim() function with logging switched on
    def _test_sim_log(self):
        """
        Test sim() function with logging switched on
        """
        # Simulate unbinned observations
        res = obsutils.sim(self._setup_sim(), log=True)

        # Check simulation results
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 4, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'EventList', 'Check event type')
        self.test_value(res[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res[0].events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')

        # Check energy dispersion flag
        self.test_assert(not (res[0].response().use_edisp()),
                         'Check energy dispersion usage')
        res[0].response().apply_edisp(True)
        self.test_assert(res[0].response().use_edisp(),
                         'Check energy dispersion usage')

        # Return
        return

    # Test set_obs() function
    def _test_set_obs(self):
        """
        Test set_obs() function
        """
        # Setup pointing direction
        pnt = gammalib.GSkyDir()
        pnt.radec_deg(83.63, 22.51)

        # Setup one CTA observation
        res = obsutils.set_obs(pnt, emin=1.0, emax=10.0)

        # Check result
        self.test_value(res.eventtype(), 'EventList', 'Check event type')
        self.test_value(res.events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res.events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')
        self.test_value(res.pointing().dir().ra_deg(), 83.63,
                        'Check pointing Right Ascension')
        self.test_value(res.pointing().dir().dec_deg(), 22.51,
                        'Check pointing declination')

        # Setup one CTA observation for local caldb directory
        res = obsutils.set_obs(pnt, caldb=self._datadir, irf='irf_file.fits',
                               emin=1.0, emax=10.0)

        # Check result
        self.test_value(res.eventtype(), 'EventList', 'Check event type')
        self.test_value(res.events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res.events().ebounds().emax().TeV(), 10.0,
                        'Check maximum energy')
        self.test_value(res.pointing().dir().ra_deg(), 83.63,
                        'Check pointing Right Ascension')
        self.test_value(res.pointing().dir().dec_deg(), 22.51,
                        'Check pointing declination')

        # Return
        return

    # Test set_obs_list() function
    def _test_set_obs_list(self):
        """
        Test set_obs_list() function
        """
        # Setup observation definition list
        obsdeflist = [{'ra': 83.63, 'dec': 22.51}]

        # Setup observation container
        res = obsutils.set_obs_list(obsdeflist)

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 0, 'Check number of models')
        self.test_value(res.nobserved(), 0, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'EventList', 'Check event type')
        self.test_value(res[0].pointing().dir().ra_deg(), 83.63,
                        'Check pointing Right Ascension')
        self.test_value(res[0].pointing().dir().dec_deg(), 22.51,
                        'Check pointing declination')

        # Return
        return

    # Test set_obs_patterns() function
    def _test_set_obs_patterns(self):
        """
        Test set_obs_pattern() function
        """
        # Get single observation pattern
        res = obsutils.set_obs_patterns('single')
        self.test_value(len(res), 1, 'Check number of observations for '
                        '"single" pattern')

        # Get four observation pattern
        res = obsutils.set_obs_patterns('four')
        self.test_value(len(res), 4, 'Check number of observations for "four"'
                        ' pattern')

        # Get invalid observations pattern
        self.test_try('Check invalid pointing pattern')
        try:
            res = obsutils.set_obs_patterns('five')
            self.test_try_failure('Exception not thrown')
        except RuntimeError:
            self.test_try_success()

        # Return
        return

    # Test set_observations() function
    def _test_set_observations(self):
        """
        Test set_observations() function
        """
        # Set observation
        res = obsutils.set_observations(83.63, 22.51, 5.0,
                                        0.0, 1800.0,
                                        1.0, 100.0,
                                        'South_50h', 'prod2',
                                        pattern='single')

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 0, 'Check number of models')
        self.test_value(res.nobserved(), 0, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_value(res[0].eventtype(), 'EventList', 'Check event type')
        self.test_value(res[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(res[0].events().ebounds().emax().TeV(), 100.0,
                        'Check maximum energy')
        self.test_value(res[0].pointing().dir().ra_deg(), 83.63,
                        'Check pointing Right Ascension')
        self.test_value(res[0].pointing().dir().dec_deg(), 22.51,
                        'Check pointing declination')

        # Return
        return

    # Test get_stacked_response() function
    def _test_get_stacked_response(self):
        """
        Test get_stacked_response() function
        """
        # Set-up observation container
        obs = self._setup_sim()

        # Set-up counts cube
        map     = gammalib.GSkyMap('CAR','CEL',83.6331,22.0145,0.1,0.1,10,10,5)
        emin    = gammalib.GEnergy(0.1, 'TeV')
        emax    = gammalib.GEnergy(100.0, 'TeV')
        ebds    = gammalib.GEbounds(5, emin, emax)
        tmin    = gammalib.GTime(0.0)
        tmax    = gammalib.GTime(1000.0)
        gti     = gammalib.GGti(tmin, tmax)
        cntcube = gammalib.GCTAEventCube(map, ebds, gti)

        # Get stacked response
        res = obsutils.get_stacked_response(obs, cntcube, edisp=False)

        # Check result
        self.test_value(res['expcube'].cube().npix(), 100,
                        'Check number of exposure cube pixels')
        self.test_value(res['expcube'].cube().nmaps(), 91,
                        'Check number of exposure cube maps')
        self.test_value(res['psfcube'].cube().npix(), 4,
                        'Check number of PSF cube pixels')
        self.test_value(res['psfcube'].cube().nmaps(), 18200,
                        'Check number of PSF cube maps')
        self.test_value(res['bkgcube'].cube().npix(), 100,
                        'Check number of background cube pixels')
        self.test_value(res['bkgcube'].cube().nmaps(), 5,
                        'Check number of background cube maps')

        # Get stacked response with energy dispersion
        res = obsutils.get_stacked_response(obs, cntcube, edisp=True)

        # Check result
        self.test_value(res['expcube'].cube().npix(), 100,
                        'Check number of exposure cube pixels')
        self.test_value(res['expcube'].cube().nmaps(), 105,
                        'Check number of exposure cube maps')
        self.test_value(res['psfcube'].cube().npix(), 4,
                        'Check number of PSF cube pixels')
        self.test_value(res['psfcube'].cube().nmaps(), 21000,
                        'Check number of PSF cube maps')
        self.test_value(res['bkgcube'].cube().npix(), 100,
                        'Check number of background cube pixels')
        self.test_value(res['bkgcube'].cube().nmaps(), 5,
                        'Check number of background cube maps')
        self.test_value(res['edispcube'].cube().npix(), 4,
                        'Check number of energy dispersion cube pixels')
        self.test_value(res['edispcube'].cube().nmaps(), 10500,
                        'Check number of energy dispersion cube maps')

        # Set-up counts cube with large number of energy bins
        ebds    = gammalib.GEbounds(100, emin, emax)
        cntcube = gammalib.GCTAEventCube(map, ebds, gti)

        # Get stacked response with xref/yref set and large number of energy bins
        res = obsutils.get_stacked_response(obs, cntcube, edisp=False)

        # Check result
        self.test_value(res['expcube'].cube().npix(), 100,
                        'Check number of exposure cube pixels')
        self.test_value(res['expcube'].cube().nmaps(), 91,
                        'Check number of exposure cube maps')
        self.test_value(res['psfcube'].cube().npix(), 4,
                        'Check number of PSF cube pixels')
        self.test_value(res['psfcube'].cube().nmaps(), 18200,
                        'Check number of PSF cube maps')
        self.test_value(res['bkgcube'].cube().npix(), 100,
                        'Check number of background cube pixels')
        self.test_value(res['bkgcube'].cube().nmaps(), 100,
                        'Check number of background cube maps')

        # Return
        return

    # Test get_stacked_obs() function
    def _test_get_stacked_obs(self):
        """
        Test get_stacked_obs() function
        """
        # Set-up unbinned cslightcrv (is not run here!!!)
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'LIN'
        lcrv['tmin']     = '2020-01-01T00:00:00'
        lcrv['tmax']     = '2020-01-01T00:05:00'
        lcrv['tbins']    = 2
        lcrv['method']   = '3D'
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['enumbins'] = 3
        lcrv['coordsys'] = 'CEL'
        lcrv['proj']     = 'TAN'
        lcrv['xref']     = 83.63
        lcrv['yref']     = 22.01
        lcrv['binsz']    = 0.1
        lcrv['nxpix']    = 10
        lcrv['nypix']    = 10
        lcrv['outfile']  = 'obsutils_cslightcrv_py1.fits'
        lcrv['logfile']  = 'obsutils_cslightcrv_py1.log'
        lcrv._get_parameters()

        # Get stacked observation container
        res = obsutils.get_stacked_obs(lcrv, lcrv.obs())

        # Apply energy dispersion
        res[0].response().apply_edisp(True)

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 96, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_assert(not res[0].response().use_edisp(),
                         'Check that no energy dispersion is used')

        # Set energy dispersion
        lcrv['edisp'] = True

        # Get stacked observation container
        res = obsutils.get_stacked_obs(lcrv, lcrv.obs())

        # Apply energy dispersion
        res[0].response().apply_edisp(True)

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 96, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_assert(res[0].response().use_edisp(),
                         'Check that energy dispersion is used')

        # Increase chatter level
        lcrv['edisp']   = False
        lcrv['chatter'] = 4

        # Get stacked observation container
        res = obsutils.get_stacked_obs(lcrv, lcrv.obs())

        # Apply energy dispersion
        res[0].response().apply_edisp(True)

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 96, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')
        self.test_assert(not res[0].response().use_edisp(),
                         'Check that no energy dispersion is used')

        # Return
        return

    # Test get_onoff_obs() function
    def _test_get_onoff_obs(self):
        """
        Test get_onoff_obs() function
        """
        # Set-up unbinned cslightcrv (is not run here!!!)
        lcrv = cscripts.cslightcrv()
        lcrv['inobs']    = self._events
        lcrv['inmodel']  = self._model
        lcrv['srcname']  = 'Crab'
        lcrv['caldb']    = self._caldb
        lcrv['irf']      = self._irf
        lcrv['tbinalg']  = 'LIN'
        lcrv['tmin']     = '2020-01-01T00:00:00'
        lcrv['tmax']     = '2020-01-01T00:05:00'
        lcrv['tbins']    = 2
        lcrv['method']   = 'ONOFF'
        lcrv['emin']     = 1.0
        lcrv['emax']     = 100.0
        lcrv['enumbins'] = 2
        lcrv['coordsys'] = 'CEL'
        lcrv['xref']     = 83.63
        lcrv['yref']     = 22.01
        lcrv['rad']      = 0.2
        lcrv['outfile']  = 'obsutils_cslightcrv_py2.fits'
        lcrv['logfile']  = 'obsutils_cslightcrv_py2.log'
        lcrv._get_parameters()

        # Get on/off observation container
        res = obsutils.get_onoff_obs(lcrv, lcrv.obs())

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 91, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')

        # Use galactic coordinates
        lcrv['coordsys'] = 'GAL'
        lcrv['xref']     = 184.5597
        lcrv['yref']     =  -5.7892
        lcrv['chatter']  = 4

        # Get on/off observation container
        res = obsutils.get_onoff_obs(lcrv, lcrv.obs())

        # Check result
        self.test_value(res.size(), 1, 'Check number of observations')
        self.test_value(res.models().size(), 2, 'Check number of models')
        self.test_value(res.nobserved(), 91, 'Check number of observed events')
        self.test_value(res.npred(), 0.0, 'Check number of predicted events')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Shows butterfly diagram created with ctbutterfly
#
# Copyright (C) 2014-2017 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ============== #
# Plot butterfly #
# ============== #
def plot_butterfly(filename, plotfile):
    """
    Plot butterfly diagram

    Parameters
    ----------
    filename : str
        Butterfly CSV file
    plotfile : str
        Plot filename
    """
    # Read butterfly file
    csv = gammalib.GCsv(filename)

    # Initialise arrays to be filled
    butterfly_x = []
    butterfly_y = []
    line_x      = []
    line_y      = []

    # Loop over rows of the file
    nrows = csv.nrows()
    for row in range(nrows):

        # Get conversion coefficient
        conv = csv.real(row,0) * csv.real(row,0) * gammalib.MeV2erg

        # Compute upper edge of confidence band
        butterfly_x.append(csv.real(row,0)/1.0e6) # TeV
        butterfly_y.append(csv.real(row,2)*conv)

        # Set line values
        line_x.append(csv.real(row,0)/1.0e6) # TeV
        line_y.append(csv.real(row,1)*conv)

    # Loop over the rows backwards to compute the lower edge of the
    # confidence band
    for row in range(nrows):
        index = nrows - 1 - row
        conv  = csv.real(index,0) * csv.real(index,0) * gammalib.MeV2erg
        butterfly_x.append(csv.real(index,0)/1.0e6)
        low_error = max(csv.real(index,3)*conv, 1e-26)
        butterfly_y.append(low_error)
    
    # Plot the butterfly and spectral line       
    plt.figure()
    plt.loglog()
    plt.grid()
    plt.plot(line_x,line_y,color='black',ls='-')
    plt.fill(butterfly_x,butterfly_y,color='green',alpha=0.5)
    plt.xlabel('Energy (TeV)')
    plt.ylabel(r'E$^2$ $\times$ dN/dE (erg cm$^{-2}$ s$^{-1}$)')

    # Show spectrum or save it into file
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ====================== #
# Show butterfly diagram #
# ====================== #
def show_butterfly():
    """
    Show butterfly diagram
    """
    # Set usage string
    usage = 'show_butterfly.py [-p plotfile] file'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Plot butterfly diagram
    plot_butterfly(args[0], plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show butterfly diagram
    show_butterfly()
